/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file AttributeMenu.cpp
\brief BMenu that lists attributes as seen in File Types.
*/

#include <NodeInfo.h>
#include <MenuItem.h>
#include <Message.h>
#include "AttributeMenu.h"

AttributeMenu::AttributeMenu(const char *name):
        BMenu(name, B_ITEMS_IN_COLUMN)
{
}


/**
 * Fill menu with items representing File Types attributes.
 *
 * All existing items are first removed and freed.
 * The list is made of two parts, divided by a separator.
 * - Attributes associated with a particular type.
 * - Attributes associated with the supertype (like 'image/').
 *
 * Invocation messages for menu items are simply those returned
 * by BNodeInfo::GetAttrInfo(), except the command.
 * Attribute public names are used as labels.
 */
status_t AttributeMenu::SetTo(const entry_ref *ref, int32 command)
{
	for (int i = CountItems()-1; i >= 0; i--)
		delete RemoveItem(i);
	
	if (!ref)
		return B_ERROR;

    // Fill Attribute Names menu.
    BNode node(ref);
    status_t retcode = node.InitCheck();
    if (retcode == B_OK) {
        BNodeInfo nodeInfo(&node);
        char mime[B_MIME_TYPE_LENGTH];
        if (nodeInfo.GetType(mime) == B_OK) {
            // Attributes associated with this particular type.
            BMimeType mimeType(mime);
            BMessage info;
            mimeType.GetAttrInfo(&info);
            const char *publicname, *name;
            int32 type;
            for (int i = 0; info.FindString("attr:public_name", i, &publicname) == B_OK; i++) {
                BMessage *msg = new BMessage(command);
                if (info.FindString("attr:name", i, &name) == B_OK)
                    msg->AddString("attr:name", name);
                if (info.FindInt32("attr:type", i, &type) == B_OK)
                    msg->AddInt32("attr:type", type);
                AddItem(new BMenuItem(publicname, msg));
            }
            // Attributes associated with the supertype.
            BMimeType mimeSupertype;
            mimeType.GetSupertype(&mimeSupertype);
            if (strcmp(mimeSupertype.Type(),"application")) {
				if (!info.IsEmpty())
                	AddSeparatorItem();
            	info.MakeEmpty();
                mimeSupertype.GetAttrInfo(&info);
                for (int i = 0; info.FindString("attr:public_name", i, &publicname) == B_OK; i++) {
                    BMessage *msg = new BMessage(command);
                    if (info.FindString("attr:name", i, &name) == B_OK)
                        msg->AddString("attr:name", name);
                    if (info.FindInt32("attr:type", i, &type) == B_OK)
                        msg->AddInt32("attr:type", type);
                    AddItem(new BMenuItem(publicname, msg));
                }
            }
        }
    }
    return retcode;
}






