/**
Copyright (c) 2005-2006 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file BrowserWindow.cpp
\author Matjaž Kovač
\brief Main Document Window
*/

#define DEBUG 1
#include <Debug.h>
#include "App.h"
#include "BrowserWindow.h"
#include "SplitView.h"
#include <stdlib.h>
#include <MenuBar.h>
#include <ScrollView.h>
#include <MenuItem.h>
#include <TextView.h>
#include <TranslatorFormats.h>
#include <FindDirectory.h>
#include <Directory.h>


BrowserWindow::BrowserWindow(BRect frame, const char *title):
	BWindow(frame, title, B_DOCUMENT_WINDOW, B_ASYNCHRONOUS_CONTROLS),
	fWriteAttr("IPRO:thumbnail")
{
	SetSizeLimits(360, 1600, 380, 1400);
    // Menu Bar
    BMenuBar *menuBar = new BMenuBar(BRect(), "MenuBar");
    AddChild(menuBar);
	
    // File Menu
    BMenu *fileMenu = new BMenu(_("File"));
    menuBar->AddItem(fileMenu);
    BMenuItem* itemAbout = new BMenuItem(_("About")B_UTF8_ELLIPSIS, new BMessage(B_ABOUT_REQUESTED));
    itemAbout->SetTarget(be_app);
    fileMenu->AddItem(itemAbout);
    fFileTrash = new BMenuItem(_("Move to Trash"), new BMessage(CMD_FILE_TRASH));
	fFileTrash->SetEnabled(false);
    fileMenu->AddItem(fFileTrash);
    fileMenu->AddSeparatorItem();
    fileMenu->AddItem(new BMenuItem(_("Quit"), new BMessage(B_QUIT_REQUESTED), 'Q'));

    // Edit Menu
    BMenu *editMenu = new BMenu(_("Edit"));
    menuBar->AddItem(editMenu);
    BMenuItem* itemSelectAll = new BMenuItem(_("Select All"), new BMessage(B_SELECT_ALL), 'A');
    editMenu->AddItem(itemSelectAll);
    editMenu->AddItem(new BMenuItem(_("Select Marked"), new BMessage(CMD_SELECT_MARKED), 'R'));
	editMenu->AddSeparatorItem();
    editMenu->AddItem(new BMenuItem(_("Mark Selected"), new BMessage(CMD_MARK_SELECTED), 'M'));
    editMenu->AddItem(new BMenuItem(_("Unmark All"), new BMessage(CMD_UNMARK_ALL), 'U'));
	editMenu->AddSeparatorItem();
    editMenu->AddItem(new BMenuItem(_("Remove Selected"), new BMessage(CMD_REMOVE_SELECTED)));

    // View Menu
    BMenu *viewMenu = new BMenu(_("View"));
    menuBar->AddItem(viewMenu);

	// View Columns Submenu
    BMenu *menuArrange = new BMenu("Columns");
    viewMenu->AddItem(menuArrange);
    menuArrange->SetRadioMode(true);
    BMessage *msg = new BMessage(CMD_VIEW_COLUMNS);
    msg->AddInt32("columns", 5);
    menuArrange->AddItem(new BMenuItem("5", msg));
    msg = new BMessage(CMD_VIEW_COLUMNS);
    msg->AddInt32("columns", 10);
    menuArrange->AddItem(new BMenuItem("10", msg));
    msg = new BMessage(CMD_VIEW_COLUMNS);
    msg->AddInt32("columns", 0);
    BMenuItem *item = new  BMenuItem("Auto", msg);
    item->SetMarked(true);
    menuArrange->AddItem(item);

	// View Sort Submenu
    BMenu *menuSort = new BMenu(_("Sort by"));
    viewMenu->AddItem(menuSort);
    menuSort->SetRadioMode(true);
    item = new BMenuItem(_("None"), new BMessage(CMD_SORT_NONE));
	item->SetMarked(true);
    menuSort->AddItem(item);
    menuSort->AddItem(new BMenuItem(_("File Name"), new BMessage(CMD_SORT_NAME)));
    menuSort->AddItem(new BMenuItem(_("File Size"), new BMessage(CMD_SORT_SIZE)));
    menuSort->AddItem(new BMenuItem(_("Last Modified"), new BMessage(CMD_SORT_DATE)));
    
	// View Labels Submenu
    BMenu *menuShow = new BMenu(_("Labels"));
    viewMenu->AddItem(menuShow);
	fShowName = new BMenuItem(_("File Name"), new BMessage(CMD_VIEW_LABEL_NAME));
    menuShow->AddItem(fShowName);
	fShowName->SetMarked(true);
	fShowSize = new BMenuItem(_("File Size"), new BMessage(CMD_VIEW_LABEL_SIZE));
    menuShow->AddItem(fShowSize);
	fShowIndicators = new BMenuItem(_("Tag Indicators"), new BMessage(CMD_VIEW_INDICATORS));
	fShowIndicators->SetMarked(true);
    viewMenu->AddItem(fShowIndicators);
	fShowFlicker = new BMenuItem(_("Flicker-free"), new BMessage(CMD_VIEW_FLICKER));
	fShowFlicker->SetMarked(true);
    viewMenu->AddItem(fShowFlicker);
	viewMenu->AddSeparatorItem();
	item = new BMenuItem(_("Preferences"), new BMessage(CMD_SHOW_PREFS));
	item->SetTarget(be_app);
    viewMenu->AddItem(item);
    
	
	// Tools Menu
    fToolsMenu = new BMenu(_("Tools"));
	menuBar->AddItem(fToolsMenu);
    fToolsMenu->AddItem(new BMenuItem(_("Make Thumbnail"), new BMessage(CMD_WRITE_THUMB)));
    fToolsMenu->AddItem(new BMenuItem(_("Make Tracker Icon"), new BMessage(CMD_WRITE_ICON)));
    fToolsMenu->AddItem(new BMenuItem(_("Delete Attributes"), new BMessage(CMD_DELETE_ATTR)));
//    fToolsMenu->AddItem(new BMenuItem(_("Add Attribute"), new BMessage(CMD_NEW_ATTR)));
//    fToolsMenu->AddItem(new BMenuItem(_("Batch Run"), new BMessage(CMD_BATCH_RUN)));
    fToolsMenu->SetEnabled(false);
	
	// Menu Status View
	BRect r = menuBar->Bounds();
	r.Set(r.right - 60, 0, r.right - 1, r.bottom - 1);
	fInfoView = new TonedStatusBar(r, "InfoView", "0", B_FOLLOW_RIGHT);
	fInfoView->SetBorderStyle(B_FANCY_BORDER);
	fInfoView->SetFontSize(11);
	menuBar->AddChild(fInfoView);

	// SplitView container
	r = Bounds();
	r.top = menuBar->Frame().bottom + 1;
	SplitView *view = new SplitView(r, "SplitView", B_VERTICAL, B_FOLLOW_ALL, B_PULSE_NEEDED | B_FRAME_EVENTS);
	
	
	// Dashboard
	fDashboard = new Dashboard(BRect(0,0,230,500), "Panel", B_FOLLOW_TOP_BOTTOM);
	view->AddChild(fDashboard);

	// Image Browser
	fBrowser = new ImageBrowser(BRect(255,0,400,500), "Browser", B_FOLLOW_ALL_SIDES);
	fBrowser->SetLabels(LABEL_NAME | LABEL_INDICATORS);
    BScrollView *scroller = new BScrollView("Scroller", fBrowser, B_FOLLOW_ALL, B_WILL_DRAW, true,true, B_NO_BORDER);
    view->AddChild(scroller);
  	fBrowser->SetBuffering(true);

	// This will auto-arrange children.
	AddChild(view);
	// Fix a visual glitch.
	scroller->ResizeBy(1,1);
	// Only attached views can be targets.
    editMenu->SetTargetForItems(fBrowser);
	// ImageBrowser needs this for updates etc.
	SetPulseRate(110000);

	// Register observers.
	fBrowser->StartWatchingAll(fDashboard);
	fBrowser->StartWatchingAll(fInfoView);
}


void BrowserWindow::MessageReceived(BMessage *message)
{
	int32 value;
	switch (message->what) {
		case MSG_ZOOM_CHANGED:
			if (message->FindInt32("be:value", &value) == B_OK)
				fBrowser->SetZoom(value / 20.0);
			break;
		case MSG_SELECTION_CHANGED:
			if (message->FindInt32("selected", &value) == B_OK) {
				fToolsMenu->SetEnabled(value);
				fFileTrash->SetEnabled(value);
			}
			break;
		case CMD_SORT_NONE:
			fBrowser->SetOrder(0);
			break;
		case CMD_SORT_NAME:
			fBrowser->SetOrder(LABEL_NAME);
			break;
		case CMD_SORT_SIZE:
			fBrowser->SetOrder(LABEL_SIZE);
			break;
		case CMD_SORT_DATE:
			fBrowser->SetOrder(LABEL_DATE);
			break;
		case CMD_VIEW_LABEL_NAME:
			fShowName->SetMarked(!fShowName->IsMarked());
			fBrowser->SetLabels(fBrowser->Labels() ^ LABEL_NAME);
			break;
		case CMD_VIEW_LABEL_SIZE:
			fShowSize->SetMarked(!fShowSize->IsMarked());
			fBrowser->SetLabels(fBrowser->Labels() ^ LABEL_SIZE);
			break;
		case CMD_VIEW_INDICATORS:
			fShowIndicators->SetMarked(!fShowIndicators->IsMarked());
			fBrowser->SetLabels(fBrowser->Labels() ^ LABEL_INDICATORS);
			break;
		case CMD_VIEW_FLICKER:
			fShowFlicker->SetMarked(!fShowFlicker->IsMarked());
			fBrowser->SetBuffering(fShowFlicker->IsMarked());
			break;
		case CMD_VIEW_COLUMNS:
			if (message->FindInt32("columns", &value) == B_OK)
				fBrowser->SetColumns(value);
			break;
		case MSG_PREFS_CHANGED:
			PrefsReceived(message);
			break;
		// Dashboard Commands
		case CMD_STOP:
			fBrowser->Loader()->Stop();
			break;
		case CMD_APPLY_TAG:
			ApplyTags(message);
			break;
		case CMD_ADD_ATTR:
			AddAttribute(message);
			break;
		case CMD_LINK_TAG:
			LinkTag(message);
			break;
		case CMD_BATCH_RENAME:
			BatchRename(message);
			break;
		case CMD_DELETE_ATTR:
			fDashboard->GetSelectedAttributes(message);
			DeleteAttributes(message);
			break;
		case CMD_WRITE_THUMB:
			WriteThumbnails();
			break;
		case CMD_WRITE_ICON:
			WriteTrackerIcons();
			break;
		case CMD_FILE_TRASH:
			MoveToTrash();
			break;
/*		case CMD_BATCH_RUN: {
			BMessage msg;
			fBrowser->GetSelection(&msg);
			BWindow *dialog = new RunShellWindow(BRect(40,40,300,200),"Batch Run", &msg);
			dialog->Show();
			break;
		}
		case CMD_NEW_ATTR: {
			BWindow *dialog = new AttributeEditWindow(BRect(40,40,300,200),"Attribute");
			dialog->Show();
			break;		
		}
*/		default:
			BWindow::MessageReceived(message);
	}
}


bool BrowserWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}



/**
	User preferencess changed, update accordingly.
	This should happen at least at startup.
*/
void BrowserWindow::PrefsReceived(BMessage *message)
{
	// Thumbnail Attribute Name
	const char *s;
	if (message->FindString("thumb_attr", &s) == B_OK) {
		sscanf(s, "%30s", fWriteAttr.LockBuffer(30));
		fWriteAttr.UnlockBuffer();
	}
	float w, h;
	if (message->FindFloat("thumb_width", &w) == B_OK &&
		message->FindFloat("thumb_height", &h) == B_OK)
		fBrowser->Loader()->SetThumbnailSize(w,h);

	int32 options = 1;
	if (message->FindInt32("load_options", &options) == B_OK)
		fBrowser->Loader()->fLoadOptions = options;
	if (message->FindMessage("tag_alias", &fBrowser->Loader()->fAliases) == B_OK);
}




/**
	Writes back tags as attributes and stuff.
		
	There are two modes of operation: 
	- 'tag_value' is specified, copy it verbatim.
	- else get tag value for each item separately and copy.
	
*/
void BrowserWindow::ApplyTags(BMessage *message)
{
	message->PrintToStream();

	BNode node;
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fBrowser->ItemAt(i)); i++) {
		if (!item->IsSelected() || node.SetTo(&item->EntryRef()) != B_OK)
			continue;
		const char *tagname;
		const char *value;
		// There may be multiple name/value pairs.
		for (int32 j = 0; message->FindString("tag_name", j, &tagname) == B_OK; j++) {
			// Menu & ListViewEditBox scenarios
			if( message->FindString("text", j, &value) == B_OK || message->FindString("tag_value", j, &value) == B_OK) {
				// Manual edit
				int32 type = 0;
				message->FindInt32("tag_type", &type);
				int32 i;
				int16 si;
				int8 b;
				float f;
				switch (type) {
					 case B_INT32_TYPE:
					 case B_UINT32_TYPE:
					 	i = atoi(value);
					 	node.WriteAttr(tagname, type, 0, &i, sizeof(i));
					 	break;
					 case B_INT16_TYPE:
					 case B_UINT16_TYPE:
					 	si = atoi(value);
					 	node.WriteAttr(tagname, type, 0, &si, sizeof(si));
					 	break;
					 case B_INT8_TYPE:
					 case B_UINT8_TYPE:
					 	b = atoi(value);
					 	node.WriteAttr(tagname, type, 0, &b, sizeof(b));
					 	break;
					 case B_FLOAT_TYPE:
						sscanf(value, "%f", &f);
						node.WriteAttr(tagname, type, 0, &f, sizeof(f));
						break;
					 case B_STRING_TYPE:
					 case B_MIME_TYPE:
						node.WriteAttr(tagname, type, 0, value, strlen(value)+1);
						break;
				}
			}
			else {
				// copy value
				ssize_t size;
				if (item->Tags().FindData(tagname, B_STRING_TYPE, (const void**)&value, &size) == B_OK)
					node.WriteAttr(tagname, B_STRING_TYPE, 0, value, size);
			}
		}
	}
}


/**
	Add a file attribute from the roster.
*/
void BrowserWindow::AddAttribute(BMessage *message)
{
	// Info from the menu selection.
    const char *name = "";
    if (message->FindString("attr:name", &name) != B_OK)
		return;
    int32 type = B_ANY_TYPE;
    message->FindInt32("attr:type", &type);
	BNode node;
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fBrowser->ItemAt(i)); i++) {
		if (!item->IsSelected())
			continue;
		node.SetTo(&item->EntryRef());
		if (node.InitCheck() != B_OK)
			continue;
		if (node.WriteAttr(name, type, 0, NULL, 0) == B_OK)
			PRINT(("'%s' added\n", name));				
	}
}

/**
	Adds a tag-attr pair to the aliases list.
*/
void BrowserWindow::LinkTag(BMessage *message)
{
	const char *tag_name;
	const char *attr_name;
	if (message->FindString("tag_name", &tag_name) == B_OK && message->FindString("attr_name", &attr_name) == B_OK) {
		fBrowser->Loader()->fAliases.RemoveName(tag_name);
		fBrowser->Loader()->fAliases.AddString(tag_name, attr_name);
		// Update User Preferences
		BMessage msg(CMD_UPDATE_PREFS);
		msg.AddMessage("tag_alias", &fBrowser->Loader()->fAliases);
	    be_app->PostMessage(&msg);
	}
}


/**
\TODO all
*/
void BrowserWindow::BatchRename(BMessage *message)
{
	// This comes from a special BAlert...
	message->PrintToStream();
	const char *value;
	if (message->FindString("text", &value) != B_OK)
		return;
	// Serialize wildcards ... 
	BString filename(value);
	filename.ReplaceFirst("*", "%d");		
	// Process each item ...
	BEntry entry;
	int n = 0;
	if (message->FindString("start_from", &value) == B_OK)
   		n = atoi(value);
   	// Clobber == overrun an existing entry when renaming
	bool clobber = false;
	message->FindBool("clobber", &clobber);
    char onename[B_FILE_NAME_LENGTH];
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fBrowser->ItemAt(i)); i++)
		if (item->IsSelected() && entry.SetTo(&item->EntryRef()) == B_OK) {
		sprintf(onename, filename.String(), n++);
		entry.Rename(onename, clobber);
	}
}

void BrowserWindow::DeleteAttributes(BMessage *message)
{
	BNode node;
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fBrowser->ItemAt(i)); i++)
		if (item->IsSelected()) {
			if (node.SetTo(&item->EntryRef()) != B_OK)
				continue;
			const char *attrname;
			int32 j = 0;
			while (message->FindString("attr_name", j++, &attrname) == B_OK) {
				node.RemoveAttr(attrname);
				PRINT(("attribute '%s' removed\n", attrname));
			}
		}
}

/**
	Generate icons with previews.
*/
void BrowserWindow::WriteTrackerIcons()
{
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fBrowser->ItemAt(i)); i++)
		if (item->IsSelected())
			item->WriteTrackerIcon();
}

/**
	Add thumbnail attr.
*/
void BrowserWindow::WriteThumbnails()
{
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fBrowser->ItemAt(i)); i++)
		if (item->IsSelected())
			item->WriteThumbnail(fWriteAttr.String(), B_PNG_FORMAT);
}

/**
	Move selected items to Trash.
*/
void BrowserWindow::MoveToTrash()
{
    // Using find_directory() is the right thing to do. Future, man!
    char trashpath[B_PATH_NAME_LENGTH] = "/boot/home/Desktop/Trash";
    find_directory(B_TRASH_DIRECTORY, 0, false, trashpath, sizeof(trashpath));
    BDirectory trash(trashpath);
    BEntry entry;
    // Must be done in reverse order because items are removed.
    ImageBrowserItem *item = NULL;
    for (int i = fBrowser->CountItems()-1; i >= 0; i--) {
        item = fBrowser->ItemAt(i);
        if (item && item->IsSelected()) {
            if (entry.SetTo(&item->EntryRef()) != B_OK)
                continue;
            // clobber==true in case it is already in the Trash.
            status_t status = entry.MoveTo(&trash, NULL, true);
            if (status != B_OK)
                PRINT(("moveto: %s\n",strerror(status)));
			// else fAlbum->RemoveItem(item);
        }
    }
}


