/**
Copyright (c) 2005-2006 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file BufferedView.cpp
\author Matjaž Kovač
\brief BufferedView class implementation
*/

#include "BufferedView.h"
#include <stdio.h>
#include <string.h>
#include <Region.h>

/**
\class BufferedView
This abstract BView descendant implements a display rendering technique
known as software double-buffering. 
The general idea is to do all the drawing in an off-screen bitmap (buffer) then 
bitblt it on the screen in a single operation.
The purpose of the whole excercise is to get rid of that annoying flicker
during animation or when rendering complex scenes.
All this is rather straight forward in the BeOS as BView objects can render to
BBitmap objects besides the normal app_server communication.

Descendants need to implement a pure virtual function DrawOffscreen() instead
of Draw() and do all their drawing through the supplied 'view'.
This makes it possible to switch to off-screen mode or the normal, 
less wasteful app_server drawing at run-time with SetBuffering().
*/

/**
	Constructs a new instance.
	
	The parameter list is essentialy the same as in BView with two additional flags:
	- 'async' set to true causes the off-screen bitmap to be drawn on screen with DrawBitmapAsync()
	instead of DrawBitmap(). 
	- If 'minUpdate' is set to true, update rectangles passed to DrawOffscreen() are clipped
	by employing GetClippingRegion(). By default BView::Draw() update rectangles sent by the app_server
	may be larger. You should read "The Interface Kit - Drawing" chapter in The Be Book to understand
	how this works.
	
	Note that there is no way to change these two settings afterwards.
	That is because they are a bit of a strategic decision and it makes no sense to do so.
	Whether adjusting them brings any noticeable benefit depends probably on your
	system's capabilities and installed drivers. 
*/

BufferedView::BufferedView(BRect rect, const char *name, uint32 mode, uint32 flags, bool async, bool minUpdate):
	BView(rect, name, mode, flags),
	fBuffering(false),
	fAsync(async),
	fMinUpdate(minUpdate),
	fBuffer(NULL),
	fBufferView(NULL)
{
}


BufferedView::~BufferedView()
{
	DeleteBuffer();
}

/**
Renders the offscreen buffer.

Off-screen method is used if IsBuffering() is true.
The offscreen buffer is allocated on demand with CreateBuffer().

This method relies on the state to be preserved in between calls!
In particular using SetViewColor() or SetLowColor() after enabling
buffering will produce undesirable results. 

\warning Do not override or use for drawing, override DrawOffscreen() instead!
*/
void BufferedView::Draw(BRect update)
{
	if (fMinUpdate) {
		BRegion clipper;
		GetClippingRegion(&clipper);
		update = update & clipper.Frame();
	}
	if (IsBuffering() && fBuffer) {
		// Prepare the off-screen buffer.
		if (fBuffer->Bounds().Width() < update.Width() || fBuffer->Bounds().Height() < update.Height())
			CreateBuffer(update.IntegerWidth(), update.IntegerHeight());
		fBuffer->Lock();
		fBufferView->ResizeTo(update.Width(), update.Height());
        fBufferView->SetOrigin(Origin() - update.LeftTop());
        
 		// Render.
        if (LowColor().alpha)
        	fBufferView->FillRect(update, B_SOLID_LOW);
		DrawOffscreen(fBufferView, update);
        // Async Blit - all pending operations must be complete!
        if (fAsync) {
			Sync();
	        DrawBitmapAsync(fBuffer, update.LeftTop());
        }
        else
	        DrawBitmap(fBuffer, update.LeftTop());
        
		// The off-screen bitmap is *being* drawn.
		fBuffer->Unlock();
	}
	else {
		DrawOffscreen(this, update);
	}
}


/**
	Allocates the off-screen bitmap.

	The problem here is how to determine the maximum view size since
	there may be multiple screens with different resolutions.
	The simplest solution is to increase the off-screen bitmap
	on demand. This is done in oversized chunks (tiles) to minimise
	memory allocation operations.
*/
void BufferedView::CreateBuffer(int32 width, int32 height)
{
	DeleteBuffer();
	int32 w = (width/BUF_W_GRAN + 1) * BUF_W_GRAN;
	int32 h = (height/BUF_H_GRAN + 1) * BUF_H_GRAN;
	fBuffer = new BBitmap(BRect(0, 0, w, h), B_RGB32, true);
	if (fBuffer) {
		fBufferView = new BView(BRect(0,0,width,height), "OffscreenBitmap", 0, 0);
		fBuffer->Lock();
        fBuffer->AddChild(fBufferView);
		fBuffer->Unlock();
	}	
}

/**
	Frees up allocated objects.
*/
void BufferedView::DeleteBuffer()
{
	if (fBuffer) {
		fBuffer->Lock();
		fBufferView->RemoveSelf();
		fBuffer->Unlock();
	    delete fBuffer;
	    delete fBufferView;
		fBuffer = NULL;
	}
}


/**
	Sets the buffering mode.
	
	If enabled, the off-screen bitmap is allocated 
	and a private view is prepared
	so it can be passed to consequent DrawOffscreen().
	Otherwise 'this' pointer is passed making DrawOffscreen()
	equivalent to BView::Draw().

	\warning Do not use SetViewColor() or SetLowColor() while enabled.

	The double-buffering technique works best with B_TRANSPARENT_COLOR 
	and is thus set automatically on enable. The original view color is saved
	as LowColor() and is used to set the off-screen view background
	in Draw(). When disabled, ViewColor() is set back to LowColor(). 
	It is essential these colors are not modified while in the buffering mode!
		
*/
void BufferedView::SetBuffering(bool enabled)
{
	fBuffering = enabled;
	if (fBuffering) {
		CreateBuffer(Bounds().IntegerWidth(), Bounds().IntegerHeight());
		SetLowColor(ViewColor());
	    SetViewColor(B_TRANSPARENT_COLOR);
		SetDrawingMode(B_OP_COPY);
	}
	else
		SetViewColor(LowColor());
}

bool BufferedView::IsBuffering()
{
	return fBuffering && fBuffer;
}


