/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file Dashboard.cpp
\author Matjaž Kovač
\brief Selected File Details Panel
*/

#define DEBUG 1
#include <Window.h>
#include <ScrollView.h>
#include <StringView.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <StringView.h>
#include <PopUpMenu.h>
#include <CheckBox.h>

#include "App.h"
#include "Dashboard.h"
#include "NameValueItem.h"
#include "ImageBrowser.h"
#include "TonedStatusBar.h"
#include "EditableListView.h"



class StatsListView : public BOutlineListView
{
	public:
	StatsListView(BRect frame, const char *name):
		BOutlineListView(frame, name, B_SINGLE_SELECTION_LIST, B_FOLLOW_ALL_SIDES, B_NAVIGABLE | B_WILL_DRAW | B_FRAME_EVENTS)
	{
	}

	virtual bool InitiateDrag(BPoint point, int32 index, bool wasSelected)
	{
	    KeyNameValueItem *item = dynamic_cast<KeyNameValueItem*>(ItemAt(index));
		// Dragging selected items reqires two clicks!
	 	if (item /*&& wasSelected*/ ) {
	           BMessage msg(CMD_APPLY_TAG);
	           msg.AddString("tag_name", item->Key());
	           msg.AddString("tag_value", item->Value());
	           DragMessage(&msg, ItemFrame(index));
			   return true;		
	    }
		return false;
	}
};




class AttributeListView : public BListView
{
	public:
	
	AttributeListView(BRect frame, const char *name):
		BListView(frame, name, B_MULTIPLE_SELECTION_LIST, B_FOLLOW_ALL, B_NAVIGABLE | B_WILL_DRAW | B_FRAME_EVENTS)
	{
	}
	
	virtual void MessageReceived(BMessage *message)
	{
		switch (message->what) {
			case CMD_APPLY_TAG:
				ApplyTag(message);
				break;
			default:
				BListView::MessageReceived(message);
				break;
		}	
	}

	virtual void MouseDown(BPoint where)
	{
	    // This is an event hook so there must be a Looper.
	    BMessage *message = Looper()->CurrentMessage();
	    int32 buttons = 0;
		message->FindInt32("buttons", &buttons);    
		// Invoke the context menu, if over an item.
		int32 i = IndexOf(where);
		if (i >=0 && buttons & B_SECONDARY_MOUSE_BUTTON) {
			// Select if none selected.
			if (!ItemAt(i)->IsSelected())
				BListView::MouseDown(where);
		    ShowContextMenu(where);
	    }
	    else
				BListView::MouseDown(where);
	    
	}
	
	virtual void MouseMoved(BPoint point, uint32 transit, const BMessage *message)
	{
		// While dragging 'message' contains source info, NULL otherwise.
		if (message != NULL && message->what == CMD_APPLY_TAG) {
	   		int32 buttons = 0;
			message->FindInt32("buttons", &buttons);    
			if (transit == B_INSIDE_VIEW && buttons) {
				int32 i = IndexOf(point);
				if (i < 0)
					DeselectAll();
				else
					Select(i);
			}
		}
		else
			BListView::MouseMoved(point, transit, message);
	}
	

	void ApplyTag(BMessage *message)
	{
		// Handle the context menu.
	    int32 buttons = 0;
		message->FindInt32("buttons", &buttons);    
		BPoint where = message->DropPoint() - BPoint(10,10);
	    BPopUpMenu *popupmenu = new BPopUpMenu("ItemOptions");
	    int32 i = CurrentSelection();
		if (i >= 0 || buttons & B_SECONDARY_MOUSE_BUTTON) {
			// Copy the sender's message.
		    BMessage *msgCopy= new BMessage(*message);
		    popupmenu->AddItem(new BMenuItem("Copy", msgCopy));
		    if (i >=0) {
		    	BMessage *msgLink= new BMessage(*message);
		    	msgLink->what = CMD_LINK_TAG;
		    	NameValueItem *item = dynamic_cast<NameValueItem*>(ItemAt(i));
		    	if (item)
		    		msgLink->AddString("attr_name", item->Name());
		    	popupmenu->AddItem(new BMenuItem("Associate", msgLink));
			}
		    popupmenu->SetTargetForItems(Window());
		    popupmenu->Go(where, true);
		    delete popupmenu;
		}
		else {		
			message->RemoveName("tag_value");
			Window()->PostMessage(message);
		}
	}
	

	/**
		Builds the view's context menu and invokes it.
	*/
	void ShowContextMenu(BPoint origin)
	{
	    BPopUpMenu *fPopUp = new BPopUpMenu("ItemOptions");
	    if (CurrentSelection() >= 0) {
		    //fPopUp->AddSeparatorItem();
		    fPopUp->AddItem(new BMenuItem(_("Delete"), new BMessage(CMD_DELETE_ATTR)));
	    }
	    // Invoke the popupmenu
	    fPopUp->SetTargetForItems(this);
		// Must be in async mode for this to work:
		fPopUp->SetAsyncAutoDestruct(true);
		ConvertToScreen(&origin);
		// Post messages, async mode - Go() returns ASAP.
	    fPopUp->Go(origin, true, false, true);
	}

	
};


class RenameDialog : public BWindow
{	
	public:
	BTextControl *fTextBox;
	BCheckBox *fClobber;
	BInvoker *fInvoker;
	
	RenameDialog(const char *title, BInvoker *invoker):
		BWindow(BRect(20, 20, 320, 140), title, B_FLOATING_WINDOW_LOOK, B_MODAL_APP_WINDOW_FEEL, B_NOT_ZOOMABLE | B_NOT_RESIZABLE),
		fInvoker(invoker)
	{
		BRect b = Bounds();
		BView *v = new BView(b, "", B_FOLLOW_ALL, 0);
		v->SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
		AddChild(v);
		
		fTextBox = new BTextControl(BRect(15, 15, 275, 40), "FromText", "Replace '*' with serials, starting with", "0", NULL);
		fTextBox->SetDivider(v->StringWidth(fTextBox->Label())+5);
		v->AddChild(fTextBox);
		
		fClobber= new BCheckBox(BRect(15,42,275,65),"Clobber", "Overrun existing entries (lose files!)", NULL);
		v->AddChild(fClobber);
		
		BButton *okButton = new BButton(BRect(0,0,50,20),"OkButton", _("Do it"), new BMessage('done'));
		okButton->ResizeToPreferred();
		b = okButton->Frame();
		okButton->MoveTo(v->Bounds().right - b.Width() - 10, v->Bounds().bottom - b.Height()-10);
		v->AddChild(okButton);

		b = okButton->Frame();
		b.OffsetBy(-b.Width()-10, 0);
		BButton *cancelButton = new BButton(b, "CancelButton", _("Cancel"), new BMessage(B_QUIT_REQUESTED));
		cancelButton->ResizeToPreferred();
		v->AddChild(cancelButton);
		
		SetDefaultButton(okButton);
	}
	
	virtual ~RenameDialog() 
	{
		delete fInvoker;
	}
	
	virtual void MessageReceived(BMessage *message) 
	{
		switch (message->what) {
			case 'done':
				if (fInvoker) {
					fInvoker->Message()->AddString("start_from", fTextBox->Text());
					fInvoker->Message()->AddBool("clobber", fClobber->Value());
					fInvoker->Invoke();
				}
				Quit();
				break;
			default:
				BWindow::MessageReceived(message);
		}
	}

};





/**
   Copy 's0' to 'out', modulo middle bytes that do not match those in 's1'.
   Mismatched chars are replaced with a single '*' wildcard.
   The idea is to have a filename matching both sources.
*/
int mergefilenames(const char *s0, const char *s1, char *out)
{
    int n0 = strlen(s0);
    int n1 = strlen(s1);
    // find matching limits
    int li = 0;
    for (li = 0; li < n0 && li < n1 && s0[li] == s1[li]; li++);
    int ri0 = n0-1;
    int ri1 = n1-1;
    while (ri0 >= li && ri1 >= li && s0[ri0] == s1[ri1]) {
        if (ri1 >= 0)
            ri1--;
        if (ri0 >= 0)
            ri0--;
    }
    // copy to dest
    if (n0 == 0)
        strcpy(out, s1);
    else if (li < n0) {
        int di = 0;
        int i;
        for (i = 0; i < li; out[di++] = s0[i++]);
        out[di++] = '*';
        for (i = ri0; i < n0; out[di++] = s0[++i]);
        out[di] = 0;
    }
    else
        strcpy(out, s0);
    return 0;
}




Dashboard::Dashboard(BRect frame, const char * name, uint32 resize, uint32 flags):
	BView(frame, name, resize, flags),
	fTagGroup0(NULL),
	fTagGroup1(NULL)
{
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));

	// Margins
    float left = 10;
    float right = Bounds().right - B_V_SCROLL_BAR_WIDTH - left;
	// Font Sensitivity 
	font_height fh;
    GetFontHeight(&fh);
    float h = fh.ascent + fh.descent + fh.leading;
    
    // Magnifier
    fScaler = new BSlider(BRect(left, 5, 100, 3*h), "Magnify", _("Scale"), new BMessage(MSG_ZOOM_CHANGED), 4, 20); 
	fScaler->SetModificationMessage(new BMessage(MSG_ZOOM_CHANGED));
	fScaler->SetHashMarks(B_HASH_MARKS_TOP);
	fScaler->SetHashMarkCount(5);
    fScaler->SetValue(20);
    fScaler->SetSnoozeAmount(10000);
	AddChild(fScaler);
    
	// Panic Button
	BRect r = fScaler->Frame().OffsetBySelf(90,0);
	r.InsetBy(10, 10);
	fStopButton = new BButton(r, "StopButton", _("Stop"), new BMessage(CMD_STOP));
	fStopButton->SetEnabled(false);
	fStopButton->ResizeToPreferred();
	AddChild(fStopButton);

	// Tag List
	AddChild(new BStringView(BRect(left, 3*h+3, right, 4.8*h-1), "Label", _("Stats & Tags")) );
    fTagList = new StatsListView(BRect(left, 5*h, right, 22*h), "TagList");
    fTagList->SetInvocationMessage(new BMessage(CMD_EDIT_ATTR));
	AddChild(new BScrollView("Scroll", fTagList, B_FOLLOW_TOP | B_FOLLOW_LEFT_RIGHT, B_WILL_DRAW | B_FRAME_EVENTS, false, true));	
  
	// Attribute List
	AddChild(new BStringView(BRect(left, 22*h+3, right, 23.8*h-1), "Label", _("Attributes")));
	r.Set(left, 24*h, right, Bounds().bottom - B_H_SCROLL_BAR_HEIGHT);
    fAttrList = new AttributeListView(r, "AttributeList");
    fAttrList->SetInvocationMessage(new BMessage(CMD_EDIT_ATTR));
	AddChild(new BScrollView("Scroll", fAttrList, B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS, false, true));	
}


/**
	Made part of the UI.
	Safe to call any BView method.
*/
void Dashboard::AttachedToWindow()
{
	fAttrList->SetTarget(this);
	fTagList->SetTarget(this);
}


void Dashboard::MessageReceived(BMessage *message)
{
 	switch (message->what){
		case B_OBSERVER_NOTICE_CHANGE:
			NoticeReceived(message);
			break;
		case CMD_EDIT_ATTR:
			ListItemInvoked(message);
			break;
		case CMD_BATCH_RENAME: {
			BString fname;
			message->FindString("text", &fname);
			if (fname.FindFirst('*') >= 0) {
				BMessage *msg = new BMessage(*message);
				RenameDialog *dlg = new RenameDialog(_("Batch Rename"), new BInvoker(msg, Looper(), Looper()));
				dlg->MoveTo(ConvertToScreen(Frame().LeftTop()) + BPoint(24,24));
				dlg->Show();
				break;
			}
			// fall-through (moves up the event handler chain)
		}
   		default:
   			BView::MessageReceived(message);
   			break;
	}
}

/**
	Augments a message with the current selection.
*/
void Dashboard::GetSelectedAttributes(BMessage *message)
{
	int32 select = -1, i = 0;
	while ((select = fAttrList->CurrentSelection(i++)) >= 0) {
		KeyNameValueItem *item = dynamic_cast<KeyNameValueItem*>(fAttrList->ItemAt(select));
		if (item) {
			message->AddString("attr_name",item->Name());
		}
	}
}


void Dashboard::NoticeReceived(BMessage *message)
{
	int32 change = 0;
	message->FindInt32("be:observe_change_what", &change);
	switch (change) {
		case MSG_SELECTION_CHANGED:
			UpdateReceived(message);
			break;
		case B_RESET_STATUS_BAR:
			fStopButton->SetEnabled(true);
			break;
		case MSG_LOADER_DONE:
			fStopButton->SetEnabled(false);
			break;
	}
}


/**
	Fill list views.
	\TODO: make it suck less
*/
void Dashboard::UpdateReceived(BMessage *message)
{
	// Preserve open groups
	if (fTagGroup0)
		fExpandGroup0 = fTagGroup0->IsExpanded();
	if (fTagGroup1)
		fExpandGroup1 = fTagGroup1->IsExpanded();
	
	
	// Clear the list - hopefully in a single update.
	for (int32 i = 0; i < fTagList->FullListCountItems(); i++)
		delete fTagList->FullListItemAt(i);
	fTagList->MakeEmpty();
	fTagGroup0 = NULL;
	fTagGroup1 = NULL;

	// Clear the list - hopefully in a single update.
	for (int32 i = 0; i < fAttrList->CountItems(); i++)
		delete fAttrList->ItemAt(i);
	fAttrList->MakeEmpty();

	// Aggregates
	char fnames[B_FILE_NAME_LENGTH] = "";
	off_t size_sum = 0L;
	time_t ctime = 0L, mtime = 0L;	

	ImageBrowserItem *item;
	int32 sel;
	int32 selGroup0 = 0, selGroup1 = 0;
	for (sel = 0; message->FindPointer("item", sel, (void**)&item) == B_OK; sel++) {
		char *name;
		type_code type;
		int32 count;

		// Attributes
		for (int32 i = 0; item->Attributes().GetInfo(B_ANY_TYPE, i, &name, &type, &count) == B_OK; i++) {
			// Search for an existing key.
			KeyNameValueItem* listitem = NULL;
			for (int32 i = 0; (listitem = (KeyNameValueItem*)fAttrList->ItemAt(i)); i++)
				if (strcmp(name, listitem->Name()) == 0) 
					break;
			// Create items.		
			if (!listitem) {
				listitem = new KeyNameValueItem(name, "");
				fAttrList->AddItem(listitem);
			}
			// Set value.
			BString oldvalue = listitem->Value();
			// Don't allow BEOS:TYPE editing etc
			//TODO: too lame, should be a property
			listitem->SetReadOnly(type != B_STRING_TYPE 
							&& type != B_INT32_TYPE && type != B_UINT32_TYPE 
							&& type != B_INT8_TYPE && type != B_UINT8_TYPE 
							&& type != B_INT16_TYPE && type != B_UINT16_TYPE
							&& type != B_FLOAT_TYPE
							);
			const void *data;
			ssize_t size;
			if (item->Attributes().FindData(name, type, &data, &size) == B_OK) {
				listitem->SetValue(type,data,size);
				if (oldvalue.FindFirst(listitem->Value()) < 0) {
					// Merge different values (shown in red).
					if (oldvalue != "")
						listitem->SetValueColor((rgb_color){255,0,0});
					oldvalue.Append(listitem->Value());
					oldvalue += " ";
				}					
			}
			listitem->SetValue(oldvalue.String());
		}
	

		// Tags
		for (int32 i = 0; item->Tags().GetInfo(B_ANY_TYPE, i, &name, &type, &count) == B_OK; i++) {
			// Search for an existing key.
			KeyNameValueItem* tagitem = NULL;
			for (int32 i = 0; i < fTagList->FullListCountItems(); i++) {
				// The list contains different types of items.
				KeyNameValueItem *li = dynamic_cast<KeyNameValueItem*>(fTagList->FullListItemAt(i)); 
				if (li && strcmp(name, li->Key()) == 0) {
					tagitem = li;
					break;
				}
			}
			// Create items/categories.		
			if (!tagitem) {
				tagitem = new KeyNameValueItem(name, "");
				tagitem->SetReadOnly(true);
				BString key = name;
				if (key.FindFirst("EXIF:") == 0) {
					if (!fTagGroup0) {
						fTagGroup0 = new BStringItem("EXIF", 0, fExpandGroup0);
						fTagList->AddItem(fTagGroup0);
					}
					fTagList->AddUnder(tagitem, fTagGroup0);
					key.Remove(0,5);
				}
				else if (key.FindFirst("IPTC:") == 0) {
					if (!fTagGroup1) 
						fTagList->AddItem(fTagGroup1 = new BStringItem("IPTC", 0, fExpandGroup1));
					fTagList->AddUnder(tagitem, fTagGroup1);
					key.Remove(0,5);
				}
				else
					fTagList->AddItem(tagitem);
				tagitem->SetName(key.String());				
			}

			BString oldvalue = tagitem->Value();
			// Set value.
			const void *data;
			ssize_t size;
			if (item->Tags().FindData(name, type, &data, &size) == B_OK) {
				tagitem->SetValue(type,data,size);
				if (oldvalue.FindFirst(tagitem->Value()) < 0) {
					// Merge different values (shown in red).
					if (oldvalue != "")
						tagitem->SetValueColor((rgb_color){255,0,0});
					oldvalue.Append(tagitem->Value());
					oldvalue += " ";
				}					
				tagitem->SetValue(oldvalue.String());
			}
			
			// some additional subtotals
			if (item->Flags() & ITEM_EXIF)
				++selGroup0;
			if (item->Flags() & ITEM_IPTC)
				++selGroup1;
			
		}
		// Aggregates
		mergefilenames(fnames, item->Name(), fnames);
		size_sum += item->Size();
		if (ctime == 0 || item->TimeCreated() < ctime) 
			ctime = item->TimeCreated();
		if (item->TimeModified() > mtime) 
			mtime = item->TimeModified();
	}
	// Show aggregated values.
	char label[48] = _("Name");
	if (sel > 1)
		sprintf(label, "%s (%d) ", _("Name"), sel);
	KeyNameValueItem *nameitem = new KeyNameValueItem(label, fnames);
	nameitem->SetKey("STAT:NAME");
	fTagList->AddItem(nameitem, 0);	

	char s[32];
	sprintf(s, "%.1f KB", (float)size_sum / 1024);
	fTagList->AddItem(new NameValueItem(_("Size"), s), 1);

	if (mtime >0) {
		NameValueItem *timeitem = new NameValueItem(_("Last Modified"), NULL);		
		timeitem->SetValue(B_TIME_TYPE, &mtime, sizeof(time_t));
		fTagList->AddItem(timeitem, 2);	
	}
	if (ctime >0) {
		NameValueItem *timeitem = new NameValueItem(_("First Created"), NULL);		
		timeitem->SetValue(B_TIME_TYPE, &ctime, sizeof(time_t));
		fTagList->AddItem(timeitem, 2);	
	}

	BMessage msg(MSG_SELECTION_CHANGED);
	msg.AddInt32("selected", sel );
	Window()->PostMessage(&msg);
}



void Dashboard::ListItemInvoked(BMessage *message)
{
	BListView *list = NULL;
	message->FindPointer("source", (void**)&list);
	if (list) {
		int32 index = -1;
		message->FindInt32("index", &index);
		KeyNameValueItem *item = dynamic_cast<KeyNameValueItem*>(list->ItemAt(index));
		if (item && !item->IsReadOnly()) {
			BRect frame = list->ItemFrame(index);
			frame.left += item->Divider(); 
			BMessage *msg = new BMessage(CMD_APPLY_TAG);
			msg->AddString("tag_name", item->Key());
			msg->AddInt32("tag_type", item->Type());
			frame.InsetBy(1,1);
	        ListViewEditBox *editBox = new ListViewEditBox(frame, "EditBox", item->Value(), msg);
			// Special cases.
			if (!strcmp(item->Key(),"STAT:NAME")) {
				msg->what = CMD_BATCH_RENAME;
				editBox->SetTarget(this);
			}
			else {
				GetSelectedAttributes(msg);
				editBox->SetTarget(Looper());
			}
	        list->AddChild(editBox);
       	}
	}
}





