/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file ImageBrowser.cpp
\author Matjaž Kovač
\brief Image Browser View
*/

#define DEBUG 1
#include <Debug.h>
#include <Region.h>
#include <Message.h>
#include <Window.h>
#include <ScrollBar.h>
#include <MessageRunner.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <Roster.h>

#include "App.h"
#include "ImageBrowser.h"
#include "BrowserWindow.h"
#include "LayoutPlan.h"
#include "OpenWithMenu.h"
#include "AttributeMenu.h"



// BObjectList linear search predicate
struct ItemEntryMatch : public UnaryPredicate<ImageBrowserItem> {
	public:
	entry_ref fRef;	
	ItemEntryMatch(entry_ref &ref): fRef(ref) { }
	virtual int operator() (const ImageBrowserItem *item) const
	{
		return item->EntryRef() != fRef;
	}
};

// BObjectList linear search predicate
struct ItemPointMatch : public UnaryPredicate<ImageBrowserItem> {
	public:
	BPoint fPoint;	
	ItemPointMatch(BPoint point): fPoint(point) { }
	virtual int operator() (const ImageBrowserItem *item) const
	{
		return !const_cast<ImageBrowserItem*>(item)->Frame().Contains(fPoint);
	}
};



ImageBrowser::ImageBrowser(BRect frame, const char * name, uint32 resize):
	BufferedView(frame, name, resize, B_WILL_DRAW | B_FRAME_EVENTS | B_PULSE_NEEDED, false, true),
	// Create an owning list, 10 items per block.
	fItems(10, true),
	fLastSelected(NULL),
	fMaxWidth(0),
	fMaxHeight(0),
	fZoom(1.0),
	fSelCount(0),
	fColumns(0),
	fOrder(0),
	fLabels(0),
	fMayDrag(false),
	fDoubleClicked(false),
	fAnimate(false),
	fUpdatePending(false)
{
	fLoader = new ImageLoader("ImageLoader");
	fLoader->Run();
	fLoader->StartWatchingAll(this);
	FixupScrollbars(0,0);
}


ImageBrowser::~ImageBrowser()
{
	PRINT(("stopping '%s' ...\n", fLoader->Name()));
	fLoader->StopWatchingAll(this);
	// Try break.
	fLoader->Stop();
	// Blocks until it locks.
	fLoader->Lock();
	fLoader->Quit();
	// BLooper deleted itself.
	PRINT(("stopped.\n"));
}


void ImageBrowser::AttachedToWindow()
{
}

void ImageBrowser::FrameResized(float width, float height)
{

	if (fItems.CountItems() == 0) {
		// Show the empty view message.
		Invalidate();
	}
	else {
		BRegion dirty;
		Arrange(&dirty);
		Invalidate(dirty.Frame());
	}
}



void ImageBrowser::DrawOffscreen(BView *view, BRect update)
{
	if (fItems.CountItems() == 0) {
		const char *text = "Drop images here";
		view->SetFontSize(16.0);
		view->SetHighColor(96,96,96);
		BRect rc = Bounds();
		view->DrawString(text, BPoint((rc.Width() - view->StringWidth(text))/2, rc.Height()/2));
	}


	ImageBrowserItem *item;
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++)
		if (update.Intersects(item->Frame())) {
			item->DrawItem(view, fLabels);
		}
	// A good idea.
	view->Sync();
}


/**
	Keyboard navigation.
*/
void ImageBrowser::KeyDown(const char *bytes, int32 numBytes)
{
	if (!fLastSelected)
		return;
		
    ImageBrowserItem *next = NULL;
    int index = fItems.IndexOf(fLastSelected);

    switch (bytes[0]) {
    case B_LEFT_ARROW:
        if (index > 0)
            next = ItemAt(index-1);
        break;
    case B_RIGHT_ARROW:
        if (index >= 0 && index < fItems.CountItems()-1)
            next = ItemAt(index+1);
        break;
    case B_UP_ARROW:
		while (index > 0) {
            	next = ItemAt(--index);
            	if (next->Frame().top < fLastSelected->Frame().top 
            		&& next->Frame().left < fLastSelected->Frame().right )
            		break;
			}
    	break;
    case B_DOWN_ARROW:
		while (index >= 0 && index < fItems.CountItems()-1) {
            	next = ItemAt(++index);
            	if (next->Frame().top > fLastSelected->Frame().bottom 
            		&& next->Frame().right > fLastSelected->Frame().left )
            		break;
			}
    	break;
	case B_DELETE:
		DeleteSelected();
		break;
    default:
        BView::KeyDown(bytes, numBytes);
        break;
    }

    if (next && next != fLastSelected) {
    	int32 mods = 0;
		// This is a message hook, so there must be a CurrentMessage().
    	Window()->CurrentMessage()->FindInt32("modifiers", &mods);
        if (!(mods & B_SHIFT_KEY))
    		SelectItems(false, 0, fItems.CountItems()-1);		
    	// Select on the new position.
        fLastSelected = next;
        fLastSelected->SetSelected(true);
        Invalidate(fLastSelected->Frame());
        fUpdatePending = true;
    }
}


void ImageBrowser::MessageReceived(BMessage *message)
{
 	switch (message->what){
		case MSG_LOADER_UPDATE:
			UpdateReceived(message);
			break;
		case MSG_LOADER_DELETED:
			DeleteReceived(message);
			break;
		case CMD_MARK_SELECTED:
			MarkItems(true, true);
			break;
		case CMD_UNMARK_SELECTED:
			MarkItems(false, true);
			break;
		case CMD_SELECT_MARKED:
			SelectMarked();
			break;
		case CMD_UNMARK_ALL:
			MarkItems(false, false);
			break;
		case CMD_REMOVE_SELECTED:
			DeleteSelected();
			break;
		case CMD_INVOKE:
			ItemInvoked(message);
			break;
    	case B_REFS_RECEIVED:
		case B_SIMPLE_DATA:
			RefsReceived(message);
			break;
		case B_SELECT_ALL:
			SelectItems(true, 0, fItems.CountItems() - 1);
			break;
		case B_OBSERVER_NOTICE_CHANGE:
			NoticeReceived(message);
			break;
		default:
			BufferedView::MessageReceived(message);
	}
}


/**
	A mouse button was pressed.
	
	Checks for various button-modifer combinations.
	Detects potential double-clicks and dragging.
*/
void ImageBrowser::MouseDown(BPoint where)
{
    // This is an event hook so there must be a Looper.
    BMessage *message = Looper()->CurrentMessage();
    int32 mods = 0, clicks = 0, buttons = 0;
    message->FindInt32("modifiers", &mods);
    message->FindInt32("clicks", &clicks);
    message->FindInt32("buttons", &buttons);
    
    ImageBrowserItem *item = fItems.FindIf(ItemPointMatch(where));
    if (item) {
        // Block selection -  all items betwen the last selection and this item.
        if (mods & B_COMMAND_KEY) {
            int32 from = fLastSelected ? fItems.IndexOf(fLastSelected) : 0;
            SelectItems(true, from, fItems.IndexOf(item));
        }
        // Normal selection - SHIFT adds more.
        else if (!item->IsSelected() ) {
            if(!(mods & B_SHIFT_KEY) || buttons & B_SECONDARY_MOUSE_BUTTON)
    			SelectItems(false, 0, fItems.CountItems()-1);
            item->SetSelected(true);
            Invalidate(item->Frame());
        }
        // Normal selection - SHIFT deselects.
        else if (mods & B_SHIFT_KEY) {
            item->SetSelected(false);
            Invalidate(item->Frame());
        }
        // Setup a potential dragndrop initiation.
        fMayDrag = true;
        fLastWhere = where;
        // Double clicks are completed in MouseUp().
        fDoubleClicked = (clicks == 2 && fLastSelected == item);
    	fUpdatePending = item != fLastSelected;

		// Pop-up the context menu.    	
		if (buttons & B_SECONDARY_MOUSE_BUTTON) {
			fMayDrag = false;
			// This is async invocation, the menu will stay on until dismissed.
			ShowContextMenu(where, item);
		}
    }
    else
    	SelectItems(false, 0, fItems.CountItems()-1);

    fLastSelected = item;
	MakeFocus(true);        
}


/**
 * Mouse tracking.
 * Detect item dragging.
 */
void ImageBrowser::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	// Ignore minor jitter...
	if (transit == B_INSIDE_VIEW && fMayDrag && abs((int)(fLastWhere.x-point.x)) + abs((int)(fLastWhere.y-point.y)) >10) {
	   fMayDrag = false;
	   ItemDragged(message);
	}
}

/** 
	A mouse button was released.
	This is where double-clicks are completed.
*/
void ImageBrowser::MouseUp(BPoint where)
{
	ImageBrowserItem *item = fItems.FindIf(ItemPointMatch(where));
	if (fDoubleClicked && item && item == fLastSelected) {
		BMessage msg;
		msg.AddRef("refs", &item->EntryRef());
		ItemInvoked(&msg);
	}
	fMayDrag = false;
}


void ImageBrowser::Pulse()
{
	// Process pending updates.
	if (fUpdatePending) {
		fUpdatePending = false;
		SelectionChanged();
	}

	// Eye candy. Do not push it!
	if (fAnimate) {
		fAnimate = false;
		BRegion dirty;
		ImageBrowserItem* item;
		for (int32 i = 0; (item = fItems.ItemAt(i)); i++) {
			if (item->Intensity() > 1) {
				item->SetIntensity(item->Intensity() >> 2);
				dirty.Include(item->Frame());
				if (item->Intensity())
					fAnimate = true;
			}
		}
		Invalidate(dirty.Frame());
	}
}


bool ImageBrowser::AddItem(ImageBrowserItem *item)
{
	bool ok = false;
	if (fOrder == LABEL_NAME)
		ok = fItems.BinaryInsert(item, ImageBrowserItem::CmpName);
	else if (fOrder == LABEL_SIZE)
		ok = fItems.BinaryInsert(item, ImageBrowserItem::CmpSize);
	else if (fOrder == LABEL_DATE)
		ok = fItems.BinaryInsert(item, ImageBrowserItem::CmpTimeModified);
	else
		ok = fItems.AddItem(item);
	// delete the splash message
	if (fItems.CountItems() == 1)
		Invalidate();
	return ok;
}


void ImageBrowser::GetSelection(BMessage *message)
{
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++)
		if (item->IsSelected()) {
			message->AddPointer("item", item);
			message->AddRef("refs", &item->EntryRef());
		}
}


/**
	Selects items.
*/
void ImageBrowser::SelectItems(bool state, int32 from, int32 to)
{
	// Swap if neccessary.
	if (from > to)
    	to ^= from ^= to ^= from; // yeah, why not!
	BRegion dirty;
	ImageBrowserItem *item;
	for (int32 i = from; i <= to && (item = fItems.ItemAt(i)); i++)
		if (item->IsSelected() != state) {
			item->SetSelected(state);
			dirty.Include(item->Frame());
		}
	Invalidate(dirty.Frame());
	fUpdatePending = true;
}

/**
	
*/
void ImageBrowser::DeleteSelected()
{
	BRegion dirty;
	BMessage msg(CMD_LOADER_DELETE);
	ImageBrowserItem *item;
	// We're deleting - do a reverse walk.
	for (int32 i = fItems.CountItems()-1; (item = fItems.ItemAt(i)); i--)
		if (item->IsSelected()) {
			dirty.Include(item->Frame());
			entry_ref ref = item->EntryRef();
			if (fItems.RemoveItem(item, true))
				msg.AddRef("refs", &item->EntryRef());				
		}

	if (fItems.CountItems() == 0) {
		// Show 'empty' splash banner.
		Invalidate();
		FixupScrollbars(0, 0);
	}
	else {		
		fLoader->PostMessage(&msg);
		Arrange(&dirty);
		Invalidate(dirty.Frame());
	}
	fUpdatePending = true;
	
	BMessage notice(B_UPDATE_STATUS_BAR);
	NotifyOnProgress(&notice);
}

/**
	Marks (selected) items.
	Marking is a secondary selection method.
*/
void ImageBrowser::MarkItems(bool state, bool onlySelected)
{
	// There is a small penalty for BView::Bounds().
	BRect bounds = Bounds();
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++)
		if ((!onlySelected || item->IsSelected()) && item->IsMarked() != state) {
			item->SetMarked(state);
			if (bounds.Intersects(item->Frame()))
				Invalidate(item->Frame());
		}
}

/**
	Selects marked items.
	Marking is a secondary selection method.
*/
void ImageBrowser::SelectMarked()
{
	// There is a small penalty for BView::Bounds().
	BRect bounds = Bounds();
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++)
		if (item->IsMarked() && !item->IsSelected()) {
			fUpdatePending = true;
			item->SetSelected(true);
			if (bounds.Intersects(item->Frame())) 
				Invalidate(item->Frame());
		}
}


/**
	Sets labels.
*/
void ImageBrowser::SetColumns(int columns)
{
	if (fColumns != columns) {
		fColumns = columns;
		Arrange();
		Invalidate();
	}
}


/**
	Sets labels.
*/
void ImageBrowser::SetLabels(uint32 labels)
{
	if (fLabels != labels) {
		fLabels = labels;
		ImageBrowserItem *item;
		for (int32 i = 0; (item = fItems.ItemAt(i)); i++)
			item->Update(this, fLabels);
		Arrange();
		Invalidate();
	}
}


/**
	Sets sorting order.
*/
void ImageBrowser::SetOrder(int order)
{
	fOrder = order;
	if (fOrder == LABEL_NAME)
		fItems.SortItems(ImageBrowserItem::CmpName);
	else if (fOrder == LABEL_SIZE)
		fItems.SortItems(ImageBrowserItem::CmpSize);
	else if (fOrder == LABEL_DATE)
		fItems.SortItems(ImageBrowserItem::CmpTimeModified);
	else
		// Restore the original sequence.
		fItems.SortItems(ImageBrowserItem::CmpSerial);
	Arrange();
	Invalidate();
}


/**
	The scale slider moved.
*/
void ImageBrowser::SetZoom(float scale)
{
	fZoom = scale;
	ImageBrowserItem *item;
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++) {
		item->SetScale(fZoom);
		item->Update(this, fLabels);
	}
	Arrange();
	Invalidate();
}






/**
	Lays out items one after another.
	
	As an option, any changes (dirty rectangles) to the existing layout 
	can be stored to a supplied region. 
	The idea is to Arrange(&dirty), include any additional rectangles 
	and then Invalidate(dirty.Frame()), resulting in a single redraw.
*/
void ImageBrowser::Arrange(BRegion *dirty)
{
	BRect bounds = Bounds().OffsetToSelf(0,0);
	bounds.InsetBy(2,2);
	FlowLayout layout(bounds, fColumns);
	// The entire set must be examined.
	fMaxWidth = 0;
	fMaxHeight = 0;
	ImageBrowserItem *item;
	bounds = Bounds();
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++) {
		BRect frame = layout.Next(item->Frame());
		if (item->Frame() != frame) {
			// The element moved, need to update.
			if (bounds.Intersects(item->Frame()) && dirty && item->Frame().left >= 0)
				dirty->Include(item->Frame());
			// Reposition the element.
			item->Frame().OffsetTo(frame.LeftTop());
			if (bounds.Intersects(frame) && dirty)
				dirty->Include(frame);
		}
		if (frame.right > fMaxWidth)
			fMaxWidth = frame.right;
		if (frame.bottom > fMaxHeight)
			fMaxHeight = frame.bottom;
	}
	FixupScrollbars(fMaxWidth, fMaxHeight);
}




/**
	Files have been dropped onto the view etc.
*/
void ImageBrowser::RefsReceived(BMessage *message)
{
	BMessage msg(B_SIMPLE_DATA);
	entry_ref ref;
	for (int i=0; message->FindRef("refs", i, &ref) == B_OK; i++)
	{
		// Skip duplicates.
		if (!fItems.FindIf(ItemEntryMatch(ref)))
			msg.AddRef("refs", &ref);
	}
	fLoader->PostMessage(&msg, NULL, this);
}


void ImageBrowser::UpdateReceived(BMessage *message)
{
	// sanity check
	entry_ref ref;
	if (message->FindRef("ref", &ref) != B_OK)
		return;

	ImageBrowserItem *item = fItems.FindIf(ItemEntryMatch(ref));
	if (!item) {
		item = new ImageBrowserItem(BRect(-5,-5,0,0));
		//const_cast<entry_ref&>(item->EntryRef()) = ref;
		item->SetRef(&ref);
		if (!AddItem(item)) {
			delete item;
			return;
		}
	}

	// Exists, but may have moved
	if (message->FindRef("newref", &ref) == B_OK)
		item->SetRef(&ref);
	
	// Indicator flags
	int32 flags = 0;
	if (message->FindInt32("flags", &flags) == B_OK)
		item->SetFlags(flags);
	
	// Image Preview
	BBitmap *bitmap;
	if (message->FindPointer("bitmap", (void**)&bitmap) == B_OK) {
		item->SetBitmap(bitmap);
		item->SetScale(fZoom);
	}
	// Stats (feat. const hacks)
	off_t fsize;
	if (message->FindInt64("fsize", &fsize) == B_OK)
		const_cast<off_t&>(item->Size()) = fsize;
	const void *data;
	ssize_t size;
	if (message->FindData("ctime", B_TIME_TYPE, &data, &size) == B_OK)
		const_cast<time_t&>(item->TimeCreated()) = *(time_t*)data;
	if (message->FindData("mtime", B_TIME_TYPE, &data, &size) == B_OK)
		const_cast<time_t&>(item->TimeModified()) = *(time_t*)data;
	
	// Metadata tags
	BMessage tags;
	if (message->FindMessage("metadata", &tags) == B_OK)
		item->Tags() = tags;
	if (message->FindMessage("attrs", &tags) == B_OK)
		item->Attributes() = tags;
		
		
	
	// Update the display.
	item->SetIntensity(255);
	item->Update(this, fLabels);
	fAnimate = true;
	BRegion dirty;
	Arrange(&dirty);
	fUpdatePending = item->IsSelected();
	Invalidate(dirty.Frame());
}


/**
	Handles MSG_LOADER_DELETE.
*/
void ImageBrowser::DeleteReceived(BMessage *message)
{
	// sanity check
	entry_ref ref;
	if (message->FindRef("ref", &ref) != B_OK)
		return;
		
	BRegion dirty;		
	ImageBrowserItem *item = fItems.FindIf(ItemEntryMatch(ref));
	if (item) {
		dirty.Include(item->Frame());
		if (fItems.RemoveItem(item, true)) {
			Arrange(&dirty);
			Invalidate(dirty.Frame());

			BMessage notice(B_UPDATE_STATUS_BAR);
			NotifyOnProgress(&notice);
		}
	}
}



/**
	Tells other views what's going on.
	Augments a message with a list of selected items.
*/
void ImageBrowser::SelectionChanged()
{
	BMessage notice;
	ImageBrowserItem *item;
	fSelCount = 0;
	for (int32 i = 0; (item = fItems.ItemAt(i)); i++) 
		if (item->IsSelected()) {
			fSelCount++;
			notice.AddPointer("item", item);
		}
	notice.AddInt32("total", fItems.CountItems());
	notice.AddInt32("selected", fSelCount);
	SendNotices(MSG_SELECTION_CHANGED, &notice);
}


/**
	Prepares a drag&drop message.
	
	Non-negotiated drag'n'drop delivers a B_SIMPLE_DATA to a target.
	Most programs respond by opening the supplied files.
	Tracker wil do a file copy.
*/
void ImageBrowser::ItemDragged(const BMessage *message)
{
    // Prepare a B_SIMPLE_DATA with Tracker's refs ordering.
    BMessage dragmsg(B_SIMPLE_DATA);
    int selectedCount = 0;
    ImageBrowserItem *item;
    for (int i=0; (item = ItemAt(i)); i++)
        if (item->IsSelected()) {
        	dragmsg.AddRef("refs", &item->EntryRef());
            selectedCount++;
        }

	// Make a bitmap to drag around.
    if (fLastSelected) {
        BBitmap *bmp = new BBitmap(fLastSelected->Bitmap(), true);
        if (bmp && selectedCount > 1) {
            // Render some text onto the dragged bitmap.
            bmp->Lock();
            BView canvas(bmp->Bounds(),"canvas",0,B_WILL_DRAW);
            bmp->AddChild(&canvas);
            char buf[10];
            BPoint p = canvas.Bounds().LeftBottom() + BPoint(2,-4);
            sprintf(buf,"%d", selectedCount);
            canvas.SetDrawingMode(B_OP_OVER);
            canvas.SetFontSize(31.0);
            canvas.SetHighColor(0,0,0);
            canvas.DrawString(buf, p + BPoint(2,2));
            canvas.SetHighColor(0,255,0);
            canvas.DrawString(buf, p);
            // BViews must not belong to anybody when deleted.
            canvas.RemoveSelf();
            //delete canvas;
            bmp->Unlock();
        }
        // Start the drag, the rest will be handled by app_server.
        BPoint where = fLastWhere - fLastSelected->Frame().LeftTop();
        DragMessage(&dragmsg, bmp, B_OP_BLEND, where);
    }
}

/**
	Launches the default app associated with an item's file.
*/
void ImageBrowser::ItemInvoked(BMessage *message)
{
    entry_ref ref;
    if (message->FindRef("refs", &ref) != B_OK)
    	return;

	const char *appsig;
	if (message->FindString("app", &appsig) == B_OK && appsig) {
		// Does not check app flags if it is argv-only app. 
		// Works for most real BeOS apps.
		BMessage *msg = new BMessage(B_REFS_RECEIVED);
		msg->AddRef("refs", &ref);
		be_roster->Launch(appsig, msg);
		delete msg;
	}
	else
		// No app specified, launch the preferred app.
		be_roster->Launch(&ref);
}


/**
	Handles B_OBSERVER_NOTICE_CHANGE.
*/
void ImageBrowser::NoticeReceived(BMessage *message)
{
	int32 change = 0;
	message->FindInt32("be:observe_change_what", &change);
	switch (change) {
		case MSG_LOADER_PROGRESS: {
			int32 done = 0, total = 0; 
			message->FindInt32("done", &done);
			message->FindInt32("total", &total);
			BMessage msg(B_UPDATE_STATUS_BAR);
			msg.AddFloat("max_value", total);
			if (done > 1 && done < total) 		
				msg.AddFloat("delta", 1);
			else {
				msg.what = B_RESET_STATUS_BAR;
			}
			NotifyOnProgress(&msg);
			break;
		}
		case MSG_LOADER_UPDATE:
			UpdateReceived(message);
			break;
		case MSG_LOADER_DELETED:
			DeleteReceived(message);
			break;
		case MSG_LOADER_DONE:
			SendNotices(MSG_LOADER_DONE);
			break;
	}
}


/**
	Tells other views what's going on.
*/
void ImageBrowser::NotifyOnProgress(BMessage *notice)
{
	char s[8];
	sprintf(s, "%ld", fItems.CountItems());
	notice->AddString("text", s);
	SendNotices(notice->what, notice);
}




/**
	Builds the view's context menu and invokes it.
*/
void ImageBrowser::ShowContextMenu(BPoint origin, ImageBrowserItem *item)
{
    BPopUpMenu *fPopUp = new BPopUpMenu("ItemOptions");
    if (item) {
        // Show Location - open the parent directory in Tracker.
        BMessage *msg;
        BEntry entry(&item->EntryRef());
        BEntry parent;
        entry.GetParent(&parent);
        entry_ref ref;

        if (parent.GetRef(&ref) == B_OK) {
            msg = new BMessage(CMD_INVOKE);
            msg->AddRef("refs", &ref);
            msg->AddString("app", "application/x-vnd.Be-TRAK");
            fPopUp->AddItem(new BMenuItem(_("Open Folder"), msg));
        }
        // Open With - launch a supporting app.
        BMenu *menu = new OpenWithMenu(_("Open With"B_UTF8_ELLIPSIS), &item->EntryRef(), CMD_INVOKE);
        menu->SetTargetForItems(this);
        fPopUp->AddItem(menu);
        fPopUp->AddSeparatorItem();

		// Check if all items are marked
		
		bool allMarked = true;
		ImageBrowserItem *qitem = NULL;
		for (int32 i = 0; (qitem = fItems.ItemAt(i)); i++)
			if (qitem->IsSelected() && !qitem->IsMarked()) {
				allMarked = false;
				break;
			}	
		if (allMarked) {
			BMenuItem *menuItem = new BMenuItem(_("Mark"), new BMessage(CMD_UNMARK_SELECTED));
			menuItem->SetMarked(true);
        	fPopUp->AddItem(menuItem);
		}
		else
        	fPopUp->AddItem(new BMenuItem(_("Mark"), new BMessage(CMD_MARK_SELECTED)));
        	
        fPopUp->AddItem(new BMenuItem(_("Remove"), new BMessage(CMD_REMOVE_SELECTED)));
        fPopUp->AddItem(new BMenuItem("Move to Trash", new BMessage(CMD_FILE_TRASH)));

        fPopUp->AddSeparatorItem();
        // Open With - launch a supporting app.
        AttributeMenu *attrmenu = new AttributeMenu(_("Add Attribute"B_UTF8_ELLIPSIS));
        attrmenu->SetTo(&item->EntryRef(), CMD_ADD_ATTR);
        attrmenu->SetTargetForItems(this);
        fPopUp->AddItem(attrmenu);
	    fPopUp->AddItem(new BMenuItem(_("Make Thumbnail"), new BMessage(CMD_WRITE_THUMB)));
	    fPopUp->AddItem(new BMenuItem(_("Make Tracker Icon"), new BMessage(CMD_WRITE_ICON)));

    }
    // Invoke the popupmenu
    fPopUp->SetTargetForItems(this);
	// Must be in async mode for this to work:
	fPopUp->SetAsyncAutoDestruct(true);
	ConvertToScreen(&origin);
	// Post messages, async mode - Go() returns ASAP.
    fPopUp->Go(origin, true, false, true);
}




/**
 * Adjusts scrollbar positions.
 * Based on "Scrolling to Oblivion" Be Newsletter article.
 */
void ImageBrowser::FixupScrollbars(float maxWidth, float maxHeight)
{
	

    BRect bounds;
    BScrollBar *sb;

    bounds = Bounds();
    float myPixelWidth = bounds.Width();
    float myPixelHeight = bounds.Height();

    float propW = myPixelWidth / maxWidth;
    float propH = myPixelHeight / maxHeight;

    float rangeW = maxWidth - myPixelWidth;
    float rangeH = maxHeight - myPixelHeight;

    float posY = 0;
    float posX = 0;

    if (rangeW < 0) rangeW = 0;
    if (rangeH < 0) rangeH = 0;

    if ((sb=ScrollBar(B_HORIZONTAL)) != NULL) {
        posX = sb->Value();

        sb->SetProportion(propW);
        sb->SetRange(0, rangeW);
        // Steps are 1/8 visible window for small steps
        //   and 1/2 visible window for large steps
        sb->SetSteps(myPixelWidth / 8.0, myPixelWidth / 2.0);
    }

    if ((sb=ScrollBar(B_VERTICAL)) != NULL) {
        posY = sb->Value();

        sb->SetProportion(propH);
        sb->SetRange(0, rangeH);
        // Steps are 1/16 visible window for small steps
        //   and 1/2 visible window for large steps
        sb->SetSteps(myPixelHeight / 16.0, myPixelHeight / 2.0);
    }
}



