#include "ImageBrowserItem.h"

#define DEBUG 1
#include <Debug.h>
#include <View.h>
#include <String.h>
#include <TranslationKit.h>
#include <NodeInfo.h>
#include <FindDirectory.h>
#include <Entry.h>
#include <Path.h>

#include "ColorTools.h"
#include "bitmap_bits.h"

BBitmap* ImageBrowserItem::fIndicatorBmap[3] = {NULL,NULL,NULL};

ImageBrowserItem::ImageBrowserItem(BRect frame):
	fBitmap(NULL),
	fFrame(frame),
	fBackColor((rgb_color){255,255,255}),
	fScale(1.0),
	fFlags(0),
	fSelected(false),
	fMarked(false),
	fIntensity(0)
{
	// Init static bitmaps once and for all.
	if (!fIndicatorBmap[0]) {
		fIndicatorBmap[0] = new BBitmap(BRect(0,0,23,23), B_RGBA32);
		memcpy(fIndicatorBmap[0]->Bits(), bits_paperclip, sizeof(bits_paperclip));
	}
	if (!fIndicatorBmap[1]) {
		fIndicatorBmap[1] = new BBitmap(BRect(0,0,15,15), B_RGBA32);
		memcpy(fIndicatorBmap[1]->Bits(), bits_exif, sizeof(bits_exif));
	}
	if (!fIndicatorBmap[2]) {
		fIndicatorBmap[2] = new BBitmap(BRect(0,0,15,15), B_RGBA32);
		memcpy(fIndicatorBmap[2]->Bits(), bits_iptc, sizeof(bits_iptc));
	}

	// Used in CmpSerial as a neutral sorting order.
	static int counter = 0;
	fSerial = counter++;
}

ImageBrowserItem::~ImageBrowserItem()
{
	delete fBitmap;
	PRINT(("deleted '%s'\n", Name()));			
}


void ImageBrowserItem::DrawItem(BView *owner, uint32 mask)
{
	owner->PushState();
	
	// Do not cache ui_colors as they may change at any time.
	rgb_color back, fore;
	rgb_color glow = {255, 0, 0};
	if (IsSelected()) {
		back = ui_color(B_MENU_SELECTION_BACKGROUND_COLOR);
		fore = ui_color(B_MENU_SELECTED_ITEM_TEXT_COLOR);
	}
	else {
		back = fBackColor;
		fore = ui_color(B_MENU_ITEM_TEXT_COLOR);
	}
	back= mix_color(back, glow, Intensity());
	
	// Background
	BRect frame = Frame();
	frame.InsetBy(1, 1);
	if (IsSelected() || Intensity() > 1) {
		owner->SetDrawingMode(B_OP_COPY);
		if (Flags() & ITEM_TRASH) 
			back = tint_color(back, B_LIGHTEN_1_TINT);
		owner->SetLowColor(back);
		owner->FillRoundRect(frame, 6, 6, B_SOLID_LOW);
	}
	frame.InsetBy(4,7);
	BPoint cur = frame.LeftTop();

	// Image Preview
	if (Bitmap())
	{
		BRect dest = Bitmap()->Bounds();
		dest.right *= Scale();
		dest.bottom *= Scale();
		dest.OffsetBy(cur);
		// Support for CMAP transparency (Tracker icons).
		owner->SetDrawingMode(B_OP_OVER);
		owner->DrawBitmapAsync(Bitmap(), dest);
		cur.y += dest.Height();
		if (Flags() & ITEM_TRASH) {
			owner->SetHighColor(ui_color(B_MENU_SELECTED_ITEM_TEXT_COLOR));
			owner->FillRect(dest, B_MIXED_COLORS);
		}
	}

	// Indicators
	owner->SetDrawingMode(B_OP_ALPHA);
	if (IsMarked())	
		owner->DrawBitmapAsync(fIndicatorBmap[0], frame.LeftTop() + BPoint(-4,-8));
	if (mask & LABEL_INDICATORS) {
		if ((Flags() & ITEM_EXIF) && frame.Width() > 30)
			owner->DrawBitmapAsync(fIndicatorBmap[1], frame.LeftTop() + BPoint(16,-8));
		if ((Flags() & ITEM_IPTC) && frame.Width() > 40)
			owner->DrawBitmapAsync(fIndicatorBmap[2], frame.LeftTop() + BPoint(25,-8));
	}
	
	// Labels
	BString label;
	owner->SetFontSize(floor(9.0 + 3.0 * fScale));	
	font_height fh;
	owner->GetFontHeight(&fh);
	// OP_COPY with background LowColor() gives the fastest AA text rendering.
	owner->SetDrawingMode(B_OP_COPY);
	owner->SetLowColor(back);
	owner->SetHighColor(fore);
	if (mask & LABEL_NAME) {
		label = Name();
		owner->TruncateString(&label, B_TRUNCATE_END, frame.Width());
		float ofs = (frame.Width() - owner->StringWidth(label.String())) / 2.0;
		cur.y += fh.ascent + fh.descent;
		owner->DrawString(label.String(), cur + BPoint(ofs, 0));
	}
	if (mask & LABEL_SIZE) {
		sprintf(label.LockBuffer(20), "%.1f KB", (float)Size() / 1024);
		label.UnlockBuffer();
		owner->TruncateString(&label, B_TRUNCATE_END, frame.Width());
		float ofs = (frame.Width() - owner->StringWidth(label.String())) / 2.0;
		cur.y += fh.ascent + fh.descent;
		owner->DrawString(label.String(), cur + BPoint(ofs, 0));
	}
	owner->PopState();
}


/**
	Adjusts Frame() to accomodate the contents.
	
	The item's owner calls this after the item is modified.
	Essentially, Update() should recalculate Frame() based on the owner's current font etc.
*/
void ImageBrowserItem::Update(BView *owner, uint32 mask)
{
	if (Bitmap()) {
		Frame().right = Frame().left + Bitmap()->Bounds().Width() * Scale();
		Frame().bottom = Frame().top + Bitmap()->Bounds().Height() * Scale();
	}
	Frame().InsetBySelf(-5, -8);
	BFont font;
	owner->GetFont(&font);
	font.SetSize(floor(9.0 + 3.0 * fScale));	
	font_height fh;
	font.GetHeight(&fh);
	if (mask & LABEL_NAME)
		fFrame.bottom +=  fh.ascent;	
	if (mask & LABEL_SIZE)
		fFrame.bottom +=  fh.ascent;	
}


/**
	Sets the bitmap (image preview).
	\warning Takes ownership!
*/
void ImageBrowserItem::SetBitmap(BBitmap *bitmap)
{
	delete fBitmap;
	fBitmap = bitmap;
}

void ImageBrowserItem::SetFlags(uint32 flags)
{
	// protect the 'trash' bit
	fFlags = flags | (fFlags & ITEM_TRASH);
}

/**
	Primary selection method.
*/
void ImageBrowserItem::SetSelected(bool enable)
{
	fSelected = enable;
}


/**
	Secondary selection method.
*/
void ImageBrowserItem::SetMarked(bool enable)
{
	fMarked = enable;
}

/**
	Sets glowing color tint level.
*/
void ImageBrowserItem::SetIntensity(uint8 level)
{
	fIntensity = level;
}


/**
	The scaling factor.
	DrawItem() and Update() must honour this.
*/
void ImageBrowserItem::SetScale(float scale)
{
	fScale = scale;
}



/**
	Neutral sorting order (as they come).
	Useful for restoring the original sequence.
*/
int ImageBrowserItem::CmpSerial(const ImageBrowserItem *a, const ImageBrowserItem *b)
{
	return(a->fSerial - b->fSerial);
}

int ImageBrowserItem::CmpName(const ImageBrowserItem *a, const ImageBrowserItem *b)
{
	return(strcmp(a->Name(), b->Name()));
}

int ImageBrowserItem::CmpSize(const ImageBrowserItem *a, const ImageBrowserItem *b)
{
	return(a->Size() - b->Size());
}

int ImageBrowserItem::CmpTimeModified(const ImageBrowserItem *a, const ImageBrowserItem *b)
{
	return(a->TimeModified() - b->TimeModified());
}




/**
	Stores the current image preview bitmap in a file attribute.
	Works only on BFS volumes.
	\param attrname File attribute name
    \param type Translation Kit data type code (like 'JPEG')
*/
status_t ImageBrowserItem::WriteThumbnail(const char *attrname, uint32 type)
{
    BTranslatorRoster *roster = BTranslatorRoster::Default();
	BBitmap *tmp = Bitmap();
    BBitmapStream in(tmp);
    BMallocIO out;
    status_t retcode = roster->Translate(&in, NULL, NULL, &out, type);
    if (retcode == B_OK)
    {
        BNode node(&EntryRef());
        retcode = node.InitCheck();
        if (retcode == B_OK)
            node.WriteAttr(attrname, B_RAW_TYPE, 0, out.Buffer(), out.BufferLength());
    }
    	PRINT(("write %s: %s\n" , attrname, strerror(retcode)));
    in.DetachBitmap(&tmp);
    return retcode;
}

/**
	Store the current image preview bitmap as a Tracker Icon.
	Works only on BFS volumes.
*/
status_t ImageBrowserItem::WriteTrackerIcon()
{
    BNode node(&fEntryRef);
    status_t status = node.InitCheck();
    if (status == B_OK) {
        BNodeInfo nodeinfo(&node);
        BView view(BRect(0, 0, 31,31),"canvas", B_WILL_DRAW,0);
	
		// B_LARGE_ICON == 32	
        BBitmap *icon = new BBitmap(BRect(0, 0, B_LARGE_ICON - 1, B_LARGE_ICON - 1), B_CMAP8, true);
        icon->AddChild(&view);
        icon->Lock();
        view.DrawBitmap(Bitmap(), icon->Bounds());
        view.RemoveSelf();
        icon->Unlock();
        status = nodeinfo.SetIcon(icon, B_LARGE_ICON);
        delete icon;

		// B_MINI_ICON == 16
        icon = new BBitmap(BRect(0, 0, B_MINI_ICON - 1, B_MINI_ICON - 1), B_CMAP8, true);
        icon->AddChild(&view);
        icon->Lock();
        view.DrawBitmap(Bitmap(), icon->Bounds());
        view.RemoveSelf();
        icon->Unlock();
        status = nodeinfo.SetIcon(icon, B_MINI_ICON);
        delete icon;
    }
    PRINT(("set icon:%s\n" ,strerror(status)));
    return status;
}


/**
	Sets a new file entry reference.
*/
status_t ImageBrowserItem::SetRef(entry_ref *ref)
{
	fEntryRef = *ref;
	// Check if this reference is in the Trash dir.
   	BPath trashpath;
  	find_directory(B_TRASH_DIRECTORY, &trashpath);
	BEntry entry(ref);
	status_t status = entry.InitCheck();
	if (status != B_OK)
		return status;
	BPath path;
	entry.GetPath(&path);
	BString paths = path.Path();
	// Trash items are marked with special gfx.
	if (paths.FindFirst(trashpath.Path()) == 0)
		fFlags |= ITEM_TRASH;
	else
		fFlags &= (~ITEM_TRASH);
	return B_OK;
}

status_t ImageBrowserItem::MoveToTrash()
{
	//TODO
	return B_OK;
}
