/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file ImageLoader.cpp
\brief Image Processor

This scheme employs the Node Monitor and the notifier-observer mechanism to
keep the client views informed on the state of the file set.
Failing to handle notices is a Bad Thing because a removed file
will result in an invalid reference.

Another neat trick is to combine the Node Monitor with Live Queries. 
The BeOS API makes it really simple as the message protocols are
essentially the same.  That way any client view can reflect 
a criterion set by a query, adding and removing files automatically as 
they come in and out of the scope of a query predicate.
*/

#define DEBUG 1
#include <Debug.h>
#include <NodeMonitor.h>
#include <NodeInfo.h>
#include <Volume.h>
#include <Directory.h>
#include <Path.h>
#include <fs_attr.h>
#include <TranslationKit.h>
#include <Bitmap.h>
#include <View.h>
#include <Autolock.h>

#include <String.h>

#include "ImageLoader.h"
#include "scale.h"

#include <libexif/exif-data.h>
#include <libexif/exif-entry.h>
#include <libiptcdata/iptc-data.h>
#include <libiptcdata/iptc-jpeg.h>




/** 
	Library-specific EXIF tag extraction.
	Courtesy of exif-data (LGPL).
*/
status_t read_exif(FILE *file, BMessage *tags, BBitmap **thumbnail)
{
	unsigned char buf[0x7fff];
	ssize_t bufsize = fread(buf, 1, sizeof(buf), file);
	if (bufsize == 0) {
		fprintf (stderr, "Error reading file\n");
		return B_ERROR;
	}
	ExifData *data = exif_data_new_from_data(buf, bufsize);
	if (!data) 
		return B_ERROR;
	// Read EXIF tags.
	BString ident;
	char value[256];
	int count = 0;
	for (int i = 0; i < EXIF_IFD_COUNT; i++) {
		if (i != 1 && data->ifd[i] && data->ifd[i]->count) {
			for (unsigned int j = 0; j < data->ifd[i]->count; j++) {
				ExifEntry *e = data->ifd[i]->entries[j];
				exif_entry_get_value(e, value, sizeof(value));
				ident = "EXIF:";
				ident += exif_tag_get_name(e->tag);
				tags->AddString(ident.String(), value);
				count++;
			}
		}
	}
	// Check EXIF image preview.
	if (data->data && data->size > 4) {
		PRINT(("EXIF thumbnail, %d bytes.\n", data->size));
		BMemoryIO in(data->data, data->size);
		*thumbnail = BTranslationUtils::GetBitmap(&in);
	}
	exif_data_unref(data);
	return count > 0 ? B_OK : B_ERROR;
}

/**
	Library specific IPTC metadata extraction.
	Courtesy of the 'iptcdata' library (LGPL).
*/
status_t read_iptc(FILE *file, BMessage *tags)
{
	// Probably too much.
	unsigned char buf[0x7fff];
	// Rewind.
	fseek(file, 0, SEEK_SET);
	int ps3_len = iptc_jpeg_read_ps3 (file, buf, sizeof(buf));
	if (ps3_len <= 0) 
		return B_ERROR;
	// This is a JPEG file - try parsing the IPTC chunk.
	unsigned int iptc_len = 0;
	int iptc_off = iptc_jpeg_ps3_find_iptc (buf, ps3_len, &iptc_len);
	if (iptc_off <= 0)
		return B_ERROR;
	IptcData *data = iptc_data_new_from_data(buf + iptc_off, iptc_len);
	if (!data) {
		fprintf(stderr, "Could not parse IPTC data\n");
		return B_BAD_VALUE;
	}
	// Read IPTC tags.
	BString ident;
	for (uint32 i=0; i < data->count; i++) {
		IptcDataSet * e = data->datasets[i];
		ident = "IPTC:";
		ident += iptc_tag_get_name(e->record, e->tag);
		const char *name = ident.String();
		switch (iptc_dataset_get_format (e)) {
			case IPTC_FORMAT_BYTE:
				tags->AddInt8(name, iptc_dataset_get_value(e));
				break;
			case IPTC_FORMAT_SHORT:
				tags->AddInt16(name, iptc_dataset_get_value(e));
				break;
			case IPTC_FORMAT_LONG:
				tags->AddInt32(name, iptc_dataset_get_value(e));
				break;
			case IPTC_FORMAT_BINARY:
				iptc_dataset_get_as_str (e, (char*)buf, sizeof(buf));
				tags->AddString(name, (char*)buf);
				break;
			default:
				iptc_dataset_get_data (e, buf, sizeof(buf));
				tags->AddString(name, (char*)buf);
				break;
		}
	}
	iptc_data_unref(data);
	return B_OK;
}


/**
	Read all attrs into a BMessage.
*/
status_t read_attributes(BNode *node, BMessage *attr)
{
	char attrname[B_ATTR_NAME_LENGTH];
	// We'll be reading previews only.
	char buf[256];
    while (node->GetNextAttrName(attrname) == B_OK) {
   		attr_info info;
   		if (node->GetAttrInfo(attrname, &info) != B_OK)
   			continue;
		//TODO: waste less space for attrs which can't be displayed.
		size_t size = sizeof(buf);
		ssize_t n = node->ReadAttr(attrname, info.type, 0, buf, size);

		// Unfortunately, AddData does not work for 0 sizes, fake something.
		if (n == 0) {
			n += 8;
			memset(buf, 0, n);
		}
		attr->AddData(attrname, info.type, buf, n, false);
    }

    return B_OK;
}




/**
	Creates a new node cache item.
*/
file_item::file_item(node_ref &node)
{
	nodref = node;
}


/**
	Node list compare function for binary search/insert.
*/
int file_item::cmp(const file_item *a, const file_item *b) 
{
	if (a->nodref.device == b->nodref.device)
		return a->nodref.node - b->nodref.node;
	else
		return a->nodref.device - b->nodref.device;
}


struct NodeEntryMatch : public UnaryPredicate<file_item> {
	public:
	entry_ref fRef;
	NodeEntryMatch(entry_ref &ref): fRef(ref) { }
	virtual int operator() (const file_item *item) const
	{
		return item->entref == fRef;
	}
};




/**
	Constructs a new message-driven image loader.
	Since it does some heavy-duty image processing it is best
	to keep it low priority to improve the overall responsivness.
*/
ImageLoader::ImageLoader(const char *name):
	BLooper(name, B_NORMAL_PRIORITY),
	// Create an owning list, 20 items per block.
	fItems(20, true),
	fRunning(false),
	fThumbWidth(130.0),
	fThumbHeight(90.0),
	fReadAttr("IPRO:thumbnail"),
	fLoadOptions(1)
{
}


ImageLoader::~ImageLoader()
{
	// Stop the node monitor.
    stop_watching(this);
}


void ImageLoader::MessageReceived(BMessage *message)
{
 	switch (message->what){
  		case B_SIMPLE_DATA:
  			RefsReceived(message);
   			break;
    	case B_NODE_MONITOR:
    	case B_QUERY_UPDATE:
			NodeMonitorChange(message);
	        break;
 		case CMD_LOADER_DELETE:
			DeleteReceived(message);
 			break;
   		default:
   			BLooper::MessageReceived(message);
   			break;
	}
}

/**
	Loads images and creates new items.

	Processes all files located by entry_refs in 'refs' and
	sends back a message for each file.
	The client view can tell something went wrong if it receives back 
	a message with no "item" pointer.

	After being loaded, files are registered with the Node Monitor.
	Note that the Node Monitor slots are a scarce resource,
	there is an per-application limit of 4096 slots (BeOS R5).
	Although not strictly neccessary, loading will fail after that
	just to keep things predictable.
	
	\warning Never call this directly from another thread! Post() only.
*/
void ImageLoader::RefsReceived(BMessage *message)
{
   	entry_ref ref;
	type_code type;
	message->GetInfo("refs", &type, &fTotal);
	fDone = 0;
   	fRunning = true;
	for (int i=0; IsBusy() && message->FindRef("refs", i, &ref) == B_OK; i++)
		HandleRef(&ref);
	fRunning = false;
	// All done.
	SendNotices(MSG_LOADER_PROGRESS);
	if (!fQuery.IsLive())
		SendNotices(MSG_LOADER_DONE);
}


/**
	Handles a generic entry_ref.
*/
status_t ImageLoader::HandleRef(entry_ref *ref)
{
	BEntry entry(ref, false);
	if (entry.IsDirectory())
	    return HandleDirectory(ref);
	else if (entry.IsFile())
		return HandleFile(ref);	 	
	else
		return B_ERROR;  
}


/**
	Handles a real file.
*/
status_t ImageLoader::HandleFile(entry_ref *ref)
{
    BNode node;
    
    if (node.SetTo(ref) != B_OK)
    	return B_ERROR;
    	
	node_ref noderef;
	node.GetNodeRef(&noderef);
	// MIME type is needed to proceed. 
    char s[2048];	
	if (node.ReadAttr("BEOS:TYPE", B_MIME_STRING_TYPE, 0, s, B_MIME_TYPE_LENGTH) == 0)
		return B_ERROR;
    else if (!strcmp(s, "application/x-vnd.Be-query")) {
		// Some queries work on a specific volume.
		if (node.ReadAttr("_trk/qryvol1", B_MESSAGE_TYPE, 0, s, sizeof(s)) == B_OK) {
			BMessage volmsg;
			volmsg.Unflatten(s);
			//TODO: set the right volume.
		}
		node.ReadAttr("_trk/qrystr", B_STRING_TYPE, 0, s, sizeof(s));
		// For now always use the volume the query came from.
		BVolume volume(noderef.device);
		HandleQuery(s, &volume);
    } 
    else {
		// Prepare a reply
		BMessage reply(MSG_LOADER_UPDATE);
		// Read the file.
		if (LoadFile(ref, &reply) == B_OK) {
			if (watch_node(&noderef, B_WATCH_ALL, this) == B_OK) {
				file_item *item = new file_item(noderef);
				item->entref = *ref;
				fItems.BinaryInsertUnique(item, file_item::cmp);
				PRINT(("B_WATCH_ALL\n"));
			}
			CurrentMessage()->SendReply(&reply);
			// Notify the observers on the progress.
			BMessage notice;
			notice.AddInt32("total", fTotal);
			notice.AddInt32("done",	++fDone);
			SendNotices(MSG_LOADER_PROGRESS, &notice);
    	}
	}
	return B_OK;
}


/**
	Handles a ref to a directory.
*/
status_t ImageLoader::HandleDirectory(entry_ref *ref)
{
	BDirectory dir(ref);
    entry_ref dirref;
	fTotal += dir.CountEntries()-1;
  	while(IsBusy() && dir.GetNextRef(&dirref) != B_ENTRY_NOT_FOUND) 
    	HandleRef(&dirref);
  	return B_OK;
}


/**
	Parses and runs a query.
*/
status_t ImageLoader::HandleQuery(const char *predicate, BVolume *volume)
{
    PRINT(("query predicate: %s\n",predicate));

    fQuery.Clear();
    fQuery.SetPredicate(predicate);
    fQuery.SetVolume(volume);

    // Make the query live.
    fQuery.SetTarget(this);
    fQuery.Fetch();

    // Get the static portion of the query.
	SendNotices(MSG_LOADER_PROGRESS);
    entry_ref ref;
    fRunning = true;
    while (IsBusy() && fQuery.GetNextRef(&ref) == B_OK)
        HandleRef(&ref);

    // Done for now, but there may be live updates later on.
    return B_OK;
}


/**
 * Stop image processing ASAP.
 *
 * Currently tries to break image loading in RefsReceived().
 */
void ImageLoader::Stop()
{
	BAutolock lock(fLocker);
	fRunning = false;
	if (fQuery.IsLive()) {
		fQuery.Clear();
		SendNotices(MSG_LOADER_DONE);
	}
}


bool ImageLoader::IsBusy()
{
	BAutolock lock(fLocker);
	return fRunning;
}



/**
	Reads an embedded preview in one of the designated attributes.
	/warning Do not forget to free the existing bitmap first!
*/
BBitmap* ReadThumbnail(BNode *node, const char *attrname)
{
	BBitmap *bitmap = NULL;
    // Attribute name can be a list of up to 4 candidates.
   	char atrlist[1024];
   	strcpy(atrlist, attrname);
    // Split strings in tokens.
    char *pch, *prog;
    pch = strtok_r(atrlist, ", ", &prog);
    attr_info attr;
    for (int i=0; pch; i++, pch = strtok_r(NULL, " ,",&prog)) {
        if (node->GetAttrInfo(pch, &attr) == B_OK) {
			// Translate a possible bitmap. 
            char *buf = (char*)malloc(attr.size);
            int n = node->ReadAttr(pch, attr.type, 0, buf, attr.size);
            BMemoryIO stream(buf, n);
			bitmap = BTranslationUtils::GetBitmap(&stream);
            free(buf);
        }
    }
    return bitmap;
}



/** 
	Load an image via the TranslationKit and scale it
*/
BBitmap* ImageLoader::MakeImagePreview(entry_ref *ref)
{
	BBitmap *bitmap = NULL;
    // Load the actual picture.
    BBitmap *original = BTranslationUtils::GetBitmap(ref);
    if (original) {
		BRect dst(0, 0, fThumbWidth - 1, fThumbHeight - 1);
		BRect src = original->Bounds();
		if (src.Width() > src.Height())
		  dst.bottom = dst.top + (src.Height() / src.Width() * dst.Width());
		else
		  dst.right = dst.left + (src.Width() / src.Height() * dst.Width());
  		bitmap = new BBitmap(dst, original->ColorSpace());
		if (bitmap) {
			scale(original, bitmap, dst.Width() / src.Width(), dst.Height() / src.Height());
/*
			// scale using app_server DrawBitmap()
			BBitmap*  tmp = new BBitmap(dst, original->ColorSpace(), true);
			BView*  view = new BView(dst, "", B_FOLLOW_ALL, B_WILL_DRAW);
			tmp->AddChild(view);
			tmp->Lock();
			view->DrawBitmap(original, dst);
			view->Sync();
			tmp->Unlock();
			memcpy(bitmap->Bits(), tmp->Bits(), tmp->BitsLength());
			// delete the temporary bitmap (deletes the attached view as well)
			delete tmp;
*/
		}
    	delete original;
    }
	return bitmap;
}



/**
	Read file data.
*/
status_t ImageLoader::LoadFile(entry_ref *ref, BMessage *reply, uint32 mode)
{
	BAutolock lock(fLocker);
	
	// sanity check
	BEntry entry(ref);
	status_t status;
	if ((status = entry.InitCheck()) != B_OK)
		return status;
	reply->AddRef("ref", ref);

	if (mode & LOAD_STATS) {
		// Get file stat.
		struct stat st;
		entry.GetStat(&st);
		reply->AddInt64("fsize", st.st_size); 
		reply->AddData("ctime", B_TIME_TYPE, &st.st_crtime, sizeof(time_t)); 
		reply->AddData("mtime", B_TIME_TYPE, &st.st_mtime, sizeof(time_t)); 
	}

	BNode node(ref);

	if (mode & LOAD_DATA) {	
		// Get an image preview.
		BBitmap *bitmap = NULL;
		BPath path(&entry);
		if (fLoadOptions) {
			FILE *file = fopen(path.Path(), "rb");
			if (file) {
				BMessage tags;
				int flags = 0;
				// EXIF tags & thumbnail
				if (read_exif(file, &tags, &bitmap) == B_OK)
					flags |= 1;
				// IPTC tags
				if (read_iptc(file, &tags) == B_OK)
					flags |= 2;
				fclose(file);
				reply->AddInt32("flags", flags);
				
				if (fLoadOptions & LOAD_OPT_TAGS) {
					// Cache EXIF, IPTC tags
					if (!tags.IsEmpty())
						reply->AddMessage("metadata", &tags);
					// Apply aliases
					if (fLoadOptions & LOAD_OPT_ALIAS) {
					char *name;
					uint32 type;
					for (int32 i = 0; i < fAliases.CountNames(B_STRING_TYPE); i++)
						if (fAliases.GetInfo(B_STRING_TYPE, i, &name, &type) == B_OK && tags.GetInfo(name, &type) == B_OK) {
							const char *attrname;
							if (fAliases.FindString(name, &attrname) == B_OK) {
								const void *buf;
								ssize_t size;
								// copy tag to attr
								if (tags.FindData(name, type, &buf, &size) == B_OK)
									node.WriteAttr(attrname, type, 0, buf, size);	
							}
						}
					}
				}
			}
		}
		// Cache EXIF bitmap, if available
		if (!bitmap) {
			BNode node (ref);
			bitmap = ReadThumbnail(&node, fReadAttr.String());
		}
		if (!bitmap) 
			// No embedded preview, make one from the actual image data.
			bitmap = MakeImagePreview(ref);
	    if (!bitmap) {
	    	// Still no picture. Load a Tracker-style icon.
	    	bitmap = new BBitmap(BRect(0,0,31,31), B_CMAP8);
	        BNodeInfo::GetTrackerIcon(ref, bitmap);
	   	}
	    if (bitmap) 
	    	reply->AddPointer("bitmap", bitmap);
		// Make sure the recipient takes this bitmap's ownership!	
	}

	// Cache attributes
	if (mode & LOAD_ATTRIBUTES) {
		if (node.InitCheck() == B_OK)  {
			BMessage attrs;
			read_attributes(&node, &attrs);
			if (!attrs.IsEmpty())
				reply->AddMessage("attrs", &attrs);
		}
	}
	
	
	return B_OK;
}


void ImageLoader::SetThumbnailSize(float width, float height)
{
	BAutolock lock(fLocker);
	fThumbWidth = width;
	fThumbHeight = height;
}


/**
	Handles node monitoring events.

	B_ENTRY_MOVED can mean either a file's name has changed or the file
	was moved to another folder. Moves across volumes do not happen as 
	such, but you will get a B_ENTRY_REMOVED on the monitored source volume.

	Some changes can be tracked either by watching each and every node or, 
	less wastefully, only parent directory nodes. 
	B_NODE_MONITOR messages look the same in both cases
	for B_ENTRY_MOVED and B_ENTRY_REMOVED but you cannot detect changes to
	a particular attribute or stat that way.
	
	All this is very similar to live query updates. Hint, hint. ;)
*/
void ImageLoader::NodeMonitorChange(BMessage *message)
{
    // Assume the presence of node_ref fields, this is the Node Monitor, after all.
    node_ref noderef;
    message->FindInt32("device", &noderef.device);
    message->FindInt64("node", &noderef.node);
	
	// Now see what's happened.
	int32 opcode;
	message->FindInt32("opcode", &opcode);    
	// Look up the watched node list.
	file_item *item = const_cast<file_item*>(fItems.BinarySearch(file_item(noderef), file_item::cmp));
    if (item) {
		BMessage msg(opcode);
		msg.AddRef("ref", &item->entref);
		switch (opcode) {
			case B_ENTRY_MOVED: {
				// A file was renamed or moved.
				PRINT(("B_ENTRY_MOVED\n"));
				const char *name = NULL;
				ino_t from, to;
				message->FindString("name", &name);
				message->FindInt64("from directory", &from);
				message->FindInt64("to directory", &to);
				item->entref = entry_ref(noderef.device, to, name);
				msg.AddRef("newref", &item->entref);
				SendNotices(MSG_LOADER_UPDATE, &msg);
				break;
	       }
	       case B_STAT_CHANGED: {
			   // A file's stat()ables changed, its data may be modified. 
	           PRINT(("B_STAT_CHANGED\n"));
			   if (LoadFile(&item->entref, &msg, LOAD_STATS) == B_OK)
			   		SendNotices(MSG_LOADER_UPDATE, &msg);
			   break;
		   }
	       case B_ATTR_CHANGED:
			   // A file's attribute set changed somehow.
	           PRINT(("B_ATTR_CHANGED\n"));
			   if (LoadFile(&item->entref, &msg, LOAD_ATTRIBUTES) == B_OK)
			   		SendNotices(MSG_LOADER_UPDATE, &msg);
	           break;
	       case B_ENTRY_REMOVED:
	           PRINT(("B_ENTRY_REMOVED\n"));
			   if (fItems.RemoveItem(item, true))
				   SendNotices(MSG_LOADER_DELETED, &msg);
	           break;
	    }
	}	   
    // This one can come only from a live query.
    else if (opcode == B_ENTRY_CREATED) {
        PRINT(("B_ENTRY_CREATED\n"));
        entry_ref ref;
        const char *name = NULL;
        ref.device = noderef.device;
        message->FindInt64("directory", &ref.directory);
        message->FindString("name", &name);
        ref.set_name(name);
        // Create a new item.
		BMessage reply;
		if (LoadFile(&ref, &reply) == B_OK) {
			if (watch_node(&noderef, B_WATCH_ALL, this) == B_OK) {
				file_item *item = new file_item(noderef);
				item->entref = ref;
				fItems.BinaryInsertUnique(item, file_item::cmp);
				PRINT(("B_WATCH_ALL\n"));
			}
			SendNotices(MSG_LOADER_UPDATE, &reply);
    	}
    }
}


/**
	Removes a set.
	Usually comes from a client view.
*/
void ImageLoader::DeleteReceived(BMessage *message)
{
	entry_ref ref;
	for (int i=0; message->FindRef("refs", i, &ref) == B_OK; i++) {
		file_item *item = fItems.FindIf(NodeEntryMatch(ref));
		if (item && watch_node(&item->nodref, B_STOP_WATCHING, this) == B_OK) {
	       	PRINT(("B_STOP_WATCHING\n"));
			fItems.RemoveItem(item, true);			
		}
	}
}




