/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file LayoutPlan.cpp
\brief Sequential Layout Engine implementation
*/

#include "LayoutPlan.h"


/**
\class Layout
Layout engine interface.

The idea is to create an instance of a particular Layout descendant, then
feed it element frames with FitNext(). The returned adjusted frames can then be used for
view positioning, drawing or whatever. 
*/
Layout::Layout(BRect frame):
	fFrame(frame),
	fSpacing(0,0)
{
}


/**
	Get the constraint frame.

	The constraint frame is merely a guide. Element frames may be placed out of bounds,
	but this should be explicitly hinted. 
*/
BRect& Layout::Frame()
{
	return (BRect&)fFrame;
}

BRect& Layout::Last()
{
	return (BRect&)fLast;
}


/**
	\fn virtual BRect Layout::Next(BRect frame, uint32 hint=0)
	Lays out an element frame.
	Essentially defines the layout rules of a descendant class.
	After each call the internal state may be changed. It is assumed that the engine was Reset() before the first call.
	An input rectangle is adjusted based on the general layout strategy, internal state and an optional hints.
	Lower 8 bits of the hint mask are reserved for standard ::layout_hints and should be either ignored or
	implemented as close to the defined meaning as possible.
	
	\param frame An element frame to be laid out.
	\param hint Special cases and exceptions.
	\returns Adjusted input frame copy
*/

BPoint Layout::Spacing()
{
	return fSpacing;
}

void Layout::SetSpacing(float x, float y)
{
	fSpacing.x = x;
	fSpacing.y = y;
}

void Layout::Reset()
{
	Last().Set(0,0,-1,-1);
}








/**
	Creates a new flow layout engine.
	See Next().
*/
FlowLayout::FlowLayout(BRect frame, int columns):
	Layout(frame),
	fMaxCol(columns)
{
	Reset();
}


/**
	Reset the engine state.
*/
void FlowLayout::Reset()
{
	Layout::Reset();
	fRowHeight = 0;
	fCol = 0;
}


/**
	Fit an element frame into the layout.
	
	Frames are fitted in a flowing fashion. 
	Next element is fitted right of the previous one if there is
	still room in Frame() otherwise the flow is broken and continues
	in the next line.

	The following hints are honored:		
	- LAYOUT_HINT_BREAK forces line break <b>after</b> this element.
	- LAYOUT_HINT_OVERRUN prevents line break <b>on</b>  this element if Frame() is overrun.
*/
BRect FlowLayout::Next(BRect rect, uint32 hint)
{
	if (!Last().IsValid())
		rect.OffsetTo(Frame().LeftTop());
	else {
		if ((!fMaxCol && Last().right + 1 + rect.Width() > Frame().Width() && !(hint & LAYOUT_HINT_OVERRUN)) 
			|| (fMaxCol && fCol >= fMaxCol -1)
			|| hint & LAYOUT_HINT_BREAK) {
			// line break
			rect.OffsetTo(Frame().left, Last().top + fRowHeight + 1);
			fRowHeight = 0;
			fCol = 0;
		}
		else {
			rect.OffsetTo(Last().right + 1, Last().top);
			fCol++;
		}
	}
	if (rect.Height() > fRowHeight)
		fRowHeight = rect.Height();
	Last() = rect;
	return rect;
}





StripesLayout::StripesLayout(BRect frame, uint32 mode):
	Layout(frame),
	fMode(mode)
{
}



/**
	Lays out an element frame.
	
	Frames are placed one next to the other in one row (or column). Unless hinted otherwise, 
	frames are extended down (or right) so they form a sort of stripes.
	Since the width (or height) of a stripe is defined by its input rectangle this only works
	well until Frame() is exceeded. 
		
	The following hints are honored:		
	- LAYOUT_HINT_LAST will extend 'frame' to completely fill the remaining Frame() area, leaving no more room.
	- LAYOUT_HINT_NO_RESIZE will only change offsets without resizing.
*/
BRect StripesLayout::Next(BRect rect, uint32 hint)
{
	if (fMode == B_HORIZONTAL) {
		if (!Last().IsValid())
			rect.OffsetTo(Frame().LeftTop());
		else
			rect.OffsetTo(Frame().left, Last().bottom + 1 + Spacing().y);
		if (!(hint & LAYOUT_HINT_NO_RESIZE)) {
			rect.right = Frame().right;
			if (hint & LAYOUT_HINT_LAST) 
				rect.bottom = Frame().bottom;
		}
	}
	else if (fMode == B_VERTICAL) {
		if (!Last().IsValid())
			rect.OffsetTo(Frame().LeftTop());
		else
			rect.OffsetTo(Last().right + 1 + Spacing().x, Frame().top);
		if (!(hint & LAYOUT_HINT_NO_RESIZE)) {
			rect.bottom = Frame().bottom;
			if (hint & LAYOUT_HINT_LAST) 
				rect.right = Frame().right;
		}
	}
	Last() = rect;
	return rect;
}


