/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file SettingsWindow.cpp
\author Matjaž Kovač
\brief User Preferences Dialog
*/

#include "App.h"
#include "SettingsWindow.h"
#include <PopUpMenu.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <CheckBox.h>
#include <Button.h>
#include <TextControl.h>
#include <ListView.h>
#include <stdio.h>
#include "NameValueItem.h"


SettingsWindow::SettingsWindow(BRect frame, const char *title):
        BWindow(frame, title, B_FLOATING_WINDOW_LOOK, B_NORMAL_WINDOW_FEEL,
                B_NOT_RESIZABLE | B_NOT_ZOOMABLE )
{
	// Container 
    BView *root = new BView(Bounds(), "", B_FOLLOW_ALL_SIDES, 0);
    root->SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));

	// Font Sensitivity
    font_height fh;
    root->GetFontHeight(&fh);
    float h = fh.ascent + fh.descent + fh.leading;
	BRect b = root->Bounds().InsetBySelf(12, 12);
	
    // Attributes Menu
    b.bottom = b.top + h;
    fReadAttr = new BTextControl(b, "Attributes", _("Thumbnail Attributes"), "IPRO:thumbnail GRAFX:Thumbnail", NULL);
    fReadAttr->SetDivider(root->StringWidth(fReadAttr->Label()) + 8);
    root->AddChild(fReadAttr);

    // Thumbnail Size Menu
    fSizeMenu = new BPopUpMenu(_("Sizes"));
    fSizeMenu->AddItem(new BMenuItem("96x64", NULL));
    fSizeMenu->AddItem(new BMenuItem("128x96", NULL));
    fSizeMenu->AddItem(new BMenuItem("200x160", NULL));
    fSizeMenu->SetTargetForItems(this);
    b.OffsetBy(0, 2*h);
	BMenuField *menuField = new BMenuField(b, "Dimensions", _("New Thumbnail Size"), fSizeMenu);
	menuField->SetDivider(root->StringWidth(menuField->Label()) + 8);
    root->AddChild(menuField);

	// Extract Tags
	b.OffsetBy(0, 2.5 * h);
    fExtractTags = new BCheckBox(b, "ExtractEXIF", _("Extract tags"), NULL);
	root->AddChild(fExtractTags);

	// Copy on Load
	b.OffsetBy(0, 1.5 * h);
	fCopyOnLoad = new BCheckBox(b,"CopyOnLoad",_("Copy tags to attributes, as listed below"), NULL);
	root->AddChild(fCopyOnLoad);
	b.OffsetBy(0, 2*h);
	b.bottom += 6*h;
	fAliasList = new BListView(b, "AliasList", B_MULTIPLE_SELECTION_LIST);
	root->AddChild(fAliasList);
	b.OffsetTo(b.left, b.bottom + 0.5*h);
	BButton *button = new BButton(b,"DeleteAssociation", "Delete", new BMessage(CMD_DELETE_TAG));
	button->ResizeToPreferred();
	root->AddChild(button);
	

    // Apply Button
	b.OffsetTo(b.right-80, root->Bounds().Height() - 3 * h);
    fApplyButton = new BButton(b, "OkButton", _("Done"), new BMessage(CMD_DONE));
	fApplyButton->ResizeToPreferred();
    root->AddChild(fApplyButton);
	fApplyButton->MakeDefault(true);

    AddChild(root);
}


/**
	Closes the window.
	Note that this window *must* be Quit() manually!
*/
bool SettingsWindow::QuitRequested()
{
    Hide();
    return false;
}


void SettingsWindow::MessageReceived(BMessage *message)
{
    switch (message->what) {
	    case MSG_PREFS_CHANGED:
	        Restore(message);
	        break;
	    case CMD_DONE:
	    	Save();
	        Hide();
	        break;
    	case CMD_DELETE_TAG:
    		DeleteAlias();
    		break;
	    default:
	        BWindow::MessageReceived(message);
	        break;
    }
}



/**
	Get anything useful from the passed message
	and update controls.
*/
void SettingsWindow::Restore(BMessage *message)
{
	message->PrintToStream();
    // Attribute names.
    const char *val;
    if (message->FindString("thumb_attr", &val) == B_OK) {
        fReadAttr->SetText(val);
    }
    // Mark an item with the right label - kind of lame.
    float width = 0, height = 0;
    if (message->FindFloat("thumb_width", &width) == B_OK && message->FindFloat("thumb_height", &height) == B_OK) {
        char label[16] = "";
        sprintf(label, "%0.0fx%0.0f", width, height);
        BMenuItem *item = fSizeMenu->FindItem(label);
        if (item)
            item->SetMarked(true);
    }
    // Read Mode
    int32 options = 0;
    if (message->FindInt32("load_options", &options) == B_OK) {
		if (options & 1)
        	fExtractTags->SetValue(1);
		if (options & 2)
        	fCopyOnLoad->SetValue(1);
    }

    //Read Aliases
	BMessage alias;
    if (message->FindMessage("tag_alias", &alias) == B_OK) {
 			while (fAliasList->CountItems()) 
        		delete fAliasList->RemoveItem(0L);	

  			char  *name;
  			const char *value;
   			uint32  type;
   			int32   count;
   			for (int32 i = 0;  alias.GetInfo(B_STRING_TYPE, i, &name, &type, &count) == B_OK; i++ ) {
				if (alias.FindString(name, &value) == B_OK) {
					NameValueItem *item = new NameValueItem(name, value);		
					fAliasList->AddItem(item);	
				}
 			}
    }
        
}


/**
	Packs all settings into a message and sends it.
*/
void SettingsWindow::Save()
{
	BMessage msg(CMD_UPDATE_PREFS);
	// Read Attributes
	msg.AddString("thumb_attr", fReadAttr->Text());
	// Thumbnail Dimensions (from labels, kind of lame).
	float width = 0, height = 0;
	if (fSizeMenu->FindMarked()) {
	    if (sscanf(fSizeMenu->FindMarked()->Label(), "%fx%f", &width, &height) == 2) {
	        msg.AddFloat("thumb_width", width);
	        msg.AddFloat("thumb_height", height);
	    }
	}
	// Read Mode
	int32 options = fExtractTags->Value();
	if (fCopyOnLoad->Value())
		options |= 2;
	msg.AddInt32("load_options", options);

	// Tag Aliases
	BMessage aliases;
	NameValueItem *item;
	for (int32 i = 0; (item = dynamic_cast<NameValueItem*>(fAliasList->ItemAt(i))); i++)
		aliases.AddString(item->Name(), item->Value());
	msg.AddMessage("tag_alias", &aliases);

    be_app->PostMessage(&msg);
}


void SettingsWindow::DeleteAlias()
{
		int32 selected = 0;
  			while ((selected = fAliasList->CurrentSelection(selected)) >= 0) 
      			delete fAliasList->RemoveItem(selected);
		//Save();
}