/**
Copyright (c) 2006-2008 by Matjaz Kovac

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do 
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.

\file Splitter.cpp
\author Matjaž Kovač
\brief View Splitter class implementation.
*/
#define DEBUG 1
#include <Debug.h>
#include "Splitter.h"


/**
\class Splitter
When dragged, this BControl descendant resizes its parent and the parent's next sibling 
(if there is one) mimicking the popular split view concept as seen elsewhere.
It can be attached to any view, but the views should already be arranged in
a tiled fashion for best results.
*/

/**
Constructs a new instance.

Attach a Splitter object to any BView with AddChild(). 
It will then automatically resize and anchor itself to the right 
or bottom edge, depending on the specified type.
There are currently two types: SPLITTER_RIGHT (attaches to the
right edge and does horizontal resizing) and SPLITTER_BOTTOM
(attaches to the bottom edge and does vertical resizing).
	
\param size Thickness (width or height)
\param name BView::Name()
\param type SPLITTER_RIGHT or SPLITTER_BOTTOM.	
*/
Splitter::Splitter(float size, const char *name, uint32 type):
	BControl(BRect(0,0,size-1,size-1), name, NULL, NULL, B_FOLLOW_RIGHT | B_FOLLOW_TOP_BOTTOM, B_WILL_DRAW | B_FRAME_EVENTS),
	fMode(type),
	fDragged(false),
	fMin(0),
	fSnap(5)
{
	if (fMode == SPLITTER_BOTTOM)
		SetResizingMode(B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM);
}

/**
	This view has been made part of a window view hierarchy.
	Repositions and resizes this view according top its type.
*/
void Splitter::AttachedToWindow()
{
	BControl::AttachedToWindow();
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	BRect frame = Parent()->Bounds();
	if (fMode == SPLITTER_RIGHT) {
		MoveTo(frame.right - Bounds().Width() - 1, 0);
		ResizeTo(Bounds().Width(), frame.Height());
	}
	else if (fMode == SPLITTER_BOTTOM) {
		MoveTo(0, frame.bottom - Bounds().Height() - 1);
		ResizeTo(frame.Width() , Bounds().Height());
	}
}

/**
	Draws a simple border.
*/
void Splitter::Draw(BRect frame)
{
	rgb_color hi, lo;
	if (fDragged) {
		hi = tint_color(ViewColor(), B_DARKEN_2_TINT);
		lo = tint_color(ViewColor(), B_LIGHTEN_2_TINT);
    }
    else {
		hi = tint_color(ViewColor(), B_LIGHTEN_2_TINT);
		lo = tint_color(ViewColor(), B_DARKEN_2_TINT);
    }
	BRect r = Bounds();
	r.InsetBy(2,2);
	BeginLineArray(4);
	AddLine(r.LeftTop(), BPoint(r.right, r.top), hi);
	AddLine(r.LeftTop(), BPoint(r.left, r.bottom), hi);
	AddLine(r.RightBottom(), BPoint(r.left + 1, r.bottom), lo);
	AddLine(r.RightBottom(), BPoint(r.right, r.top + 1), lo);
	EndLineArray();
	r.InsetBy(1,1);
	SetHighColor(ViewColor());
	FillRect(r);
}


void Splitter::FrameResized(float width, float height)
{
	Draw(Bounds());
}

void Splitter::FrameMoved(BPoint point)
{
}

/**
	Starts dragging.
*/
void Splitter::MouseDown(BPoint where)
{
	if (!fDragged) {
		fDragged = true;
		fGrabPoint = where;
		Invalidate();
		// Subscribe to off-view mouse events.
		SetMouseEventMask(B_POINTER_EVENTS, B_NO_POINTER_HISTORY | B_LOCK_WINDOW_FOCUS);
	}
}

/**
	Ends dragging.
*/
void Splitter::MouseUp(BPoint where)
{
	fDragged = false;
	Invalidate();
}

/**
	Tracks mouse movement.
*/
void Splitter::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
//	if (transit == B_ENTERED_VIEW || transit == B_EXITED_VIEW)
//		return;
	if (fDragged && Parent()) {
		BView *sibling = Parent()->NextSibling();
		BPoint delta = point - fGrabPoint;
		delta.x = (floor(delta.x/5)) *5;
		delta.y = (floor(delta.y/5)) *5;
		BPoint p = ConvertToParent(delta);
		if (fMode == SPLITTER_RIGHT) {
			if (p.x < fMin)
				delta.x = 0;
			Parent()->ResizeBy(delta.x, 0);
			if (sibling) {
				sibling->MoveBy(delta.x, 0);
				sibling->ResizeBy(-delta.x, 0);
			}
		}
		else if (fMode == SPLITTER_BOTTOM) {
			if (p.y < fMin)
				delta.y = 0;
			Parent()->ResizeBy(0, delta.y);
			if (sibling) {
				sibling->ResizeBy(0, -delta.y);
				sibling->MoveBy(0, delta.y);
			}
		}
	}
}

void Splitter::SetLimits(float min, float max)
{
	fMin = min;
	fMax = max;
}